var Cc = Components.classes;
var Ci = Components.interfaces;
var Cu = Components.utils;
var Cr = Components.results;
Cu.import('resource://gre/modules/Services.jsm');
Cu.import("resource://oneapmrecorder/ComponentLogger.jsm");
var getLogger = null;

function logException(ex, name){
    var msg = ex + " filename: " + ex.fileName ;
    msg += " lineNumber: " + ex.lineNumber ;
    getLogger().error("caught exception in " + name + " : " + msg);
}

// For any number of browsers route events to all registered event sinks
function OneApmBrowserEventCenter(mainWindow, actionMaxNum) {
    getLogger = ComponentLoggerFactory.getLoggerFunc("OneApmBrowserEventCenter");
    this.sources = new Array();
    getLogger().debug("init OneApmBrowserEventCenter");
    this.store = new RecorderActionStore(mainWindow, actionMaxNum);
    var observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    observerService.addObserver(this, "http-on-modify-request", false);
}

OneApmBrowserEventCenter.prototype.badCertListener = null;
OneApmBrowserEventCenter.prototype.observe = function(subject, topic, data) {
    var channel = subject.QueryInterface(Ci.nsIHttpChannel);
    if (topic == "http-on-modify-request" && this.badCertListener && channel.loadGroup) {
        channel.loadGroup.notificationCallbacks = this.badCertListener.create();
    }
    if (topic == "http-on-modify-request" && (channel.loadFlags & Ci.nsIRequest.LOAD_BACKGROUND)) {
        try {
            if(channel.notificationCallbacks) {
                var xhr = channel.notificationCallbacks.getInterface(Ci.nsIXMLHttpRequest);
                if(xhr) {
                    // Make a network event
                    var networkEvent = new OneAPMNetworkEvent();
                    networkEvent.timeStamp = new Date().getTime();
                    networkEvent.type = "statechange";

                    // Show that it is different than other network events
                    networkEvent.source = "http-on-modify-request";

                    // Set some flags.  The more the better!
                    networkEvent.isRequest = true;
                    networkEvent.isNetwork = true;
                    networkEvent.isStart = true;

                    var request = channel.QueryInterface(Ci.nsIRequest);
                    if (request) {
                        networkEvent.uri = request.name;
                    }
                    this.handleNetworkEvent(networkEvent);
                }
            }
        } catch(ex) {
            logException(ex,"observe");
        }
    }
}

OneApmBrowserEventCenter.prototype.eventNames = [
    "DOMActivate",
    "DOMFocusIn",
    "DOMFocusOut",
    "DOMTitleChanged",
    "focus",
    "blur",
    "textInput",
    "click",
    "dblclick",
    "mousedown",
    "mouseup",
    "mouseover",
    "mousemove",
    "mouseout",
    "keydown",
    "keyup",
    "keypress",
    "mousemultiwheel",
    "mousewheel",
    "DOMSubtreeModified",
    "DOMNodeInserted",
    "DOMNodeRemoved",
    "DOMNodeRemovedFromDocument",
    "DOMNodeInsertedIntoDocument",
    "DOMAttrModified",
    "DOMCharacterDataModified",
    "DOMElementNameChanged",
    "DOMAttributeNameChanged",
    "load",
    "unload",
    "abort",
    "error",
    "select",
    "change",
    "submit",
    "reset",
    "resize",
    "scroll",
    "input",
    "compositionstart",
    "compositionend"
];

// All DOM events for all browsers routed through this function.
OneApmBrowserEventCenter.prototype.handleDomEvent = function(domEvent) {
    var timestamp = new Date().getTime();
    try {
        if(this.store && this.store.handleDomEvent) {
            this.store.handleDomEvent(timestamp, domEvent);
        }
    } catch (ex) {
        logException(ex,"handleDomEvent");
    }
}

// All Network events for all browsers route through this function
OneApmBrowserEventCenter.prototype.handleNetworkEvent = function(networkEvent) {
    try {
        if(this.store && this.store.handleNetworkEvent) {
            this.store.handleNetworkEvent(networkEvent.timeStamp, networkEvent);
        }
    } catch (ex) {
        logException(ex,"handleNetworkEvent");
    }
}

OneApmBrowserEventCenter.prototype.finish = function() {
    var timestamp = new Date().getTime();

    // clean up the http observer
    var observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    observerService.removeObserver(this, "http-on-modify-request");

    // finish them (use mortal kombat voice)
    try {
        var store = this.store;
        this.store = null;
        if(store && store.finish) {
            return store.finish(timestamp);
        }

    } catch (ex) {
        logException(ex,"finish");
    }
}
// Attach to a browser
OneApmBrowserEventCenter.prototype.attach = function(browser) {
    //getLogger().debug("Attaching to browser...");

    var index = this.indexOfSource(browser);
    if(index == -1) {
        //Creates and adds a BrowserSource
        var browserSource = this.addSource(browser);

        //Create and add a webProgressListener
        browserSource.webProgressListener = new BrowserSourceWebProgressListener(this, browser);

        browser.addProgressListener(browserSource.webProgressListener, Ci.nsIWebProgress.NOTIFY_ALL);

        //Creates an event callback with a closure pointing to this OneApmBrowserEventCenter
        browserSource.createHandleEventCallback(this);

        for(var i=0;i<this.eventNames.length;i++) {
            browserSource.browser.addEventListener(this.eventNames[i], browserSource.handleEventCallback, true);
        }
    } else {
        getLogger().error("already attached (?)");
        // Already attached
    }
}

OneApmBrowserEventCenter.prototype.detach = function(browser) {
    var index = this.indexOfSource(browser);
    if(index != -1) {
        this.detachSource(this.sources[index]);
    }
}

// Detatch all browsers
OneApmBrowserEventCenter.prototype.detachAll = function() {
    var sources = this.sources.concat();
    for(var i=0;i<sources.length;i++) {
        var browserSource = sources[i];
        this.detachSource(browserSource);
    }
}

// Detatch from a browser
OneApmBrowserEventCenter.prototype.detachSource = function(browserSource) {
    this.removeSource(browserSource);
    if (browserSource && browserSource.browser) {
        if (browserSource.browser.removeProgressListener) {
          browserSource.browser.removeProgressListener(browserSource.webProgressListener);
        }
        if (browserSource.handleEventCallback) {
            for(var i=0;i<this.eventNames.length;i++) {
                browserSource.browser.removeEventListener(this.eventNames[i], browserSource.handleEventCallback, true);
            }
        }
    }
}

// Add a Browser (source).  Will not add a source that is already registered.
OneApmBrowserEventCenter.prototype.addSource = function(browser) {
    var browserSource = null;
    var index = this.indexOfSource(browser);
    if(index == -1) {
        browserSource = new BrowserSourceClosure(browser);
        this.sources.push(browserSource);
    } else {
        browserSource = this.sources[index];
    }

    return browserSource;
}

// Remove a BrowserSourceClosure (not a Browser)
OneApmBrowserEventCenter.prototype.removeSource = function(browserSource) {
    var index = this.sources.indexOf(browserSource);
    if(index != -1) {
        this.sources.splice(index,1);
    }
    return browserSource;
}

OneApmBrowserEventCenter.prototype.indexOfSource = function(browser) {
    var index = -1;
    for(var i=0;i<this.sources.length;i++) {
        var browserSource = this.sources[i];
        if(browserSource.browser === browser) {
            index = i;
            break;
        }
    }
    return index;
}

// Add a BrowserEventSink
OneApmBrowserEventCenter.prototype.addSink = function(browserEventSink) {
    if(this.sinks.indexOf(browserEventSink) == -1) {
        this.sinks.push(browserEventSink);
    }
}

// Remove a BrowserEventSink
OneApmBrowserEventCenter.prototype.removeSink = function(browserEventSink) {
    var index = this.sinks.indexOf(browserEventSink);
    if(index != -1) {
        this.sinks.splice(index,1);
    }
}

function BrowserSourceClosure(browser) {
    if(!browser) {
        throw { error: "ArgumentException", message: "A valid browser object is required to create a BrowserSource." };
    }

    this.browser = browser;
    this.handleEventCallback = null;
    this.webProgressListener = null;
}

// Uses a closure to keep the eventProcessor around for when the event is dispatched
BrowserSourceClosure.prototype.createHandleEventCallback = function(browserEventCenter) {
    var callback = function(domEvent) {
        browserEventCenter.handleDomEvent(domEvent);
        //getLogger().debug("Event type " + domEvent.type );
    };

    this.handleEventCallback = callback;
    return this.handleEventCallback;
}

function BrowserSourceWebProgressListener(browserEventCenter, browser) {
    this.browserEventCenter = browserEventCenter;
    this.browser = browser;
    this.stateIsRequest = false;
}

BrowserSourceWebProgressListener.prototype = {
    QueryInterface : function(aIID) {
        if (aIID.equals(Ci.nsIWebProgressListener) ||
            aIID.equals(Ci.nsISupportsWeakReference) ||
            aIID.equals(Ci.nsISupports))
        return this;
        throw Cr.NS_NOINTERFACE;
    },

    onStateChange : function(aProgress, aRequest, aFlag, aStatus) {
        var networkEvent = new OneAPMNetworkEvent();
        networkEvent.timeStamp = new Date().getTime();
        networkEvent.type = "statechange";
        networkEvent.isRequest = ((aFlag & Ci.nsIWebProgressListener.STATE_IS_REQUEST) != 0);
        networkEvent.isDocument = ((aFlag & Ci.nsIWebProgressListener.STATE_IS_DOCUMENT) != 0);
        networkEvent.isNetwork = ((aFlag & Ci.nsIWebProgressListener.STATE_IS_NETWORK) != 0);
        networkEvent.isWindow = ((aFlag & Ci.nsIWebProgressListener.STATE_IS_WINDOW) != 0);
        networkEvent.isStart = ((aFlag & Ci.nsIWebProgressListener.STATE_START) != 0);
        networkEvent.isRedirecting = ((aFlag & Ci.nsIWebProgressListener.STATE_REDIRECTING) != 0);
        networkEvent.isTransferring = ((aFlag & Ci.nsIWebProgressListener.STATE_TRANSFERRING) != 0);
        networkEvent.isNegotiating = ((aFlag & Ci.nsIWebProgressListener.STATE_NEGOTIATING) != 0);
        networkEvent.isStop = ((aFlag & Ci.nsIWebProgressListener.STATE_STOP) != 0);
        if (aRequest) networkEvent.uri = aRequest.name;
        // Notify the appropriate listeners
        this.browserEventCenter.handleNetworkEvent(networkEvent);
        return 0;
    },

    onLocationChange : function(aProgress, aRequest, aLocation) {
        if(aLocation.scheme.indexOf("chrome") !== -1) {
            return 0;
        }

        var networkEvent = new OneAPMNetworkEvent();
        networkEvent.timeStamp = new Date().getTime();
        networkEvent.type = "locationchange";
        //getLogger().debug("LocationChange: URI: "+aLocation.spec+"; request: "+aRequest);
        networkEvent.uri = aLocation.spec;
        if (networkEvent.uri.indexOf('http') != 0) {
            // bad uri (possibly javascript:), so we replace with current window location
            networkEvent.uri = this.browser.contentWindow.document.URL;
        }
        networkEvent.causedRequest = aRequest != null;
        // Notify the appropriate listeners
        this.browserEventCenter.handleNetworkEvent(networkEvent);

        // update location bar
        this.browser.updateUrl();

        return 0;
    },

    onProgressChange : function(aWebProgress,aRequest,aCurSelfProgress,aMaxSelfProgress,aCurTotalProgress,aMaxTotalProgress){
        var networkEvent = new OneAPMNetworkEvent();
        networkEvent.timeStamp = new Date().getTime();
        networkEvent.type = "progresschange";
        networkEvent.uri = aRequest.name;
        networkEvent.message = "self "+aCurSelfProgress + "/" + aMaxSelfProgress +"; total "+aCurTotalProgress+"/"+aMaxTotalProgress;
        // Notify the appropriate listeners
        this.browserEventCenter.handleNetworkEvent(networkEvent);
    },

    onStatusChange : function(aWebProgress, aRequest, aStatus, aMessage){
        var networkEvent = new OneAPMNetworkEvent();
        networkEvent.timeStamp = new Date().getTime();
        networkEvent.type = "statuschange";
        networkEvent.uri = aRequest.name;
        networkEvent.message = aMessage + " ("+ aRequest.name +")";
        // Notify the appropriate listeners
        this.browserEventCenter.handleNetworkEvent(networkEvent);
    },

    onSecurityChange : function(webProgress, request, state){
        var networkEvent = new OneAPMNetworkEvent();
        networkEvent.timeStamp = new Date().getTime();
        networkEvent.type = "securitychange";

        if((state & Ci.nsIWebProgressListener.STATE_IS_INSECURE) != 0) {
            networkEvent.state = "Insecure";
        } else if((state & Ci.nsIWebProgressListener.STATE_IS_SECURE) != 0) {
            networkEvent.state = "Secure";
        } else if((state & Ci.nsIWebProgressListener.STATE_IS_BROKEN) != 0) {
            networkEvent.state = "Broken";
        } else {
            networkEvent.state = "None";
        }

        if((state & Ci.nsIWebProgressListener.STATE_SECURE_HIGH) != 0) {
            networkEvent.strength = "High";
        } else if((state & Ci.nsIWebProgressListener.STATE_SECURE_MED) != 0) {
            networkEvent.strength = "Medium";
        } else if((state & Ci.nsIWebProgressListener.STATE_SECURE_LOW) != 0) {
            networkEvent.strength = "Low";
        } else {
        }

        try {
            var url = this.browser.contentWindow.document.URL,
                authMan = Cc["@mozilla.org/network/http-auth-manager;1"].getService(Ci.nsIHttpAuthManager),
                url = Services.io.newURI(url, null, null),
                userDomain = {},
                userName = {},
                userPassword = {},
                port = -1,
                parser = Services.wm.getMostRecentWindow(null).document.createElementNS("http://www.w3.org/1999/xhtml", "a");

            parser.href = url.spec;

            try { port = url.port } catch(ex) {}

            if(port == -1) {
                port = (url.scheme == 'http') ? 80 : -1;
                port = (url.scheme == 'https') ? 443 : -1;
            }

            authMan.getAuthIdentity(url.scheme, parser.hostname, port, "", "", parser.pathname, userDomain, userName, userPassword);

            networkEvent.credentials = { url: url.spec, domain: userDomain.value, user: userName.value, password: userPassword.value };

        } catch(ex) {
            //getLogger().error("auth exception that you can ignore: " + ex);
        }

        // Notify the appropriate listeners
        this.browserEventCenter.handleNetworkEvent(networkEvent);
    },

    onLinkIconAvailable : function(a){
        var networkEvent = new OneAPMNetworkEvent();
        networkEvent.timeStamp = new Date().getTime();
        networkEvent.type = "linkicon";

        // Notify the appropriate listeners
        this.browserEventCenter.handleNetworkEvent(networkEvent);
    }
}