var Cu = Components.utils;
Cu.import("resource://oneapmrecorder/ComponentLogger.jsm");
var getLogger = null;

function logException(ex, name){
    var msg = ex + " filename: " + ex.fileName ;
    msg += " lineNumber: " + ex.lineNumber ;
    getLogger().error("caught exception in " + name + " : " + msg);
}

function RecorderActionStore(mainWindow, actionMaxNum) {
    getLogger = ComponentLoggerFactory.getLoggerFunc("RecorderActionStore");
    getLogger().debug("Init RecorderActionStore");
    this.mainWindow = mainWindow;
    this.recorderPathInfo;
    this.events = new Array();
    this.script = [];

    this.lookforNetwork = false;
    this.lastKeyDown;
    this.currentStep;
    this.transaction;
    this.selectAction;
    this.lastMouseEvent;
    this.needFlushMouseEvent = false;
    this.keytinUpdate = null;
    this.isIME = false;
    this.isComposition = false;
    this.actionMaxNum = actionMaxNum;
    this.actionNum = 0;
    this.breakRecord = false;
}

RecorderActionStore.prototype.handleDomEvent = function(timestamp, domEvent) {
    domEvent = this.filterDomEvent(timestamp, domEvent);
    if(domEvent) {
        this.processDomEvent(timestamp, domEvent);
    }
}

RecorderActionStore.prototype.filterDomEvent = function(timestamp, domEvent) {
    if (/^composition(start|end)$/.test(domEvent.type.toLowerCase())){
        return domEvent;
    }

    if (/^domtitlechanged/.test(domEvent.type.toLowerCase())) {
        return domEvent;
    }
    //TODO: Don't process events we don't intend to playback
    if (/^mouse(over|out)|^dom|^blur|^focus/.test(domEvent.type.toLowerCase())) {
        //getLogger().debug("Discarding Event case 1: '" + domEvent.type + "'. This event type is ignored.");
        return null;
    }

    //TODO: Don't process events that have no target
    if (!domEvent.target) {
        //getLogger().debug("Discarding Event case 2: '" + domEvent.type + "'. Target property is empty.");
        return null;
    }

    //TODO: Don't process events targeted at the document itself
    if (domEvent.target.nodeName == "#document") {
        if (domEvent.type == "scroll") return domEvent;
        //getLogger().debug("Discarding Event case 3: '" + domEvent.type + "'. Target nodeName '" + domEvent.target.nodeName + "' is ignored.");
        return null;
    }

    //TODO: Don't proces events with no tagName
    if(!domEvent.target.tagName) {
        //getLogger().debug("Discarding Event case 4: '" + domEvent.type + "'. Target has no tagName property.");
        return null;
    }

    //TODO: Don't process events targetd at the html elements
    var tagName = domEvent.target.tagName.toLowerCase();

    //TODO: Don't process events targetd at the body elements
    if(tagName == "body") {
        //getLogger().debug("Discarding Event case 5'" + domEvent.type + "'. Target tagName '" + tagName + "' is ignored.");
        return null;
    }

    return domEvent;
}

RecorderActionStore.prototype.processDomEvent = function(timestamp, domEvent) {
    try {
        if(!domEvent) {
            //TODOMC log this
            return;
        }

        // Translate to RecDomEvent and save.  Save is mostly for debugging and can be turned off if necessary.
        var event = new RecDomEvent(timestamp, domEvent);
        this.events.push(event);
        var step = this.getCurrentStep();

        switch (event.eventType) {
            case "click":
                if(event.eventOptions.button == 0) {
                    if (event.origTarget && this.lastKeyDown &&
                        this.locatorsEqual(event.origTarget[0], this.lastKeyDown.locators[0])) {
                        var keyPressEvent = {
                            windowPath: this.lastKeyDown.windowPath,
                            targetLocators: this.lastKeyDown.targetLocators,
                            targetInfo: {
                                selectionStart: -1,
                                selectionEnd: -1,
                                textValue: ""
                            },
                            eventOptions: {
                                keyCode: this.lastKeyDown.eventOptions.keyCode,
                                charCode: this.lastKeyDown.eventOptions.keyCode,
                                ctrlKey: this.lastKeyDown.eventOptions.ctrlKey,
                                altKey: this.lastKeyDown.eventOptions.altKey,
                                shiftKey: this.lastKeyDown.eventOptions.shiftKey,
                                metaKey: this.lastKeyDown.eventOptions.metaKey
                            }
                        };
                        this.handleKeyPress(step, keyPressEvent);
                    }

                    if (this.locatorsEqual(this.lastMouseEvent.targetLocators[0], event.targetLocators[0])) {
                        if (this.needFlushMouseEvent) {
                            this.needFlushMouseEvent = false;
                        } else {
                            break;
                        }
                    } else {
                        if (this.needFlushMouseEvent)
                            this.flushMouseEvent();
                    }

                    getLogger().debug("Create a click action.");
                    this.addAction(this.createClick(event));

                    if (event.targetType == "embed" || event.targetType == "object") // ask Geoff how to get to the wmode='windowed'
                    {
                        // find a timestamp for the prior click event (if any)
                        var key;
                        var ts = this.events[0].timeStamp;
                        for (var i=0; i<this.events.length; i++)
                        {
                            var e = this.events[i];
                            if ((e.timeStamp && (e.timeStamp > ts) && (e.timeStamp < timestamp))&&
                                //TODO: remove the oneApmType
                                (e.oneApmType && e.oneApmType == "dom") &&
                                (e.eventType) && (e.eventType == "click" | e.eventType == "dblclick")) {
                                ts = e.timeStamp;
                            }
                        }

                        getLogger().debug("Create a wait for time.");
                        this.addAction(this.createWaitFor('time', (timestamp - ts + 500)));
                    }

                    // was there a previous select action?  If so, flush it.
                    this.flushSelectAction();
                    this.lookforNetwork = true;
                }
                break;
            case "mousemove":
                break;
            case "dblclick":
                if(event.eventOptions.button == 0) {
                    var action = this.createClick(event, true);
                    var prevAction2;
                    var needDeleteNum = 0;
                    do {
                        needDeleteNum++;
                        prevAction2 = step.actions[step.actions.length-needDeleteNum];
                    } while (prevAction2 && "wait" == prevAction2.type)
                    if (prevAction2 && "click" == prevAction2.type && this.locatorsEqual(prevAction2.locators[0], action.locators[0])) {
                        while (needDeleteNum) {
                            this.deleteAction();
                            needDeleteNum--;
                        }

                        var prevAction1;
                        needDeleteNum = 0;
                        do {
                            needDeleteNum++;
                            prevAction1 = step.actions[step.actions.length-needDeleteNum];
                        } while (prevAction1 && "wait" == prevAction1.type)
                        if (prevAction1 && "click" == prevAction1.type && this.locatorsEqual(prevAction1.locators[0], action.locators[0])) {
                            while (needDeleteNum) {
                                this.deleteAction();
                                needDeleteNum--;
                            }
                        }
                    }
                    getLogger().debug("Create a double click action.");
                    this.addAction(action);
                    this.flushSelectAction();
                }
                break;
            case "mousedown":
                // we have a new mousedown, if there was a previous mousedown with no click we
                // should clear it.
                this.flushMouseEvent();
                this.lastMouseEvent = event;
                this.needFlushMouseEvent = true;
                this.lookforNetwork = true;
                break;
            case "keydown":
                this.flushMouseEvent();
                // save the keycode
                this.lastKeyDown = event;
                break;
            case "keypress":
                this.handleKeyPress(step, event);
                break;
            case "keyup":
                this.flushSelectAction();
                this.lastKeyDown = null;
                break;
            case "input":
                if (this.isIME) {
                    this.handleKeyPress(step, event);
                    if (!this.isComposition) {
                        this.isIME = false;
                    }
                }
                if (this.keytinUpdate) this.keytinUpdate();
                break;
            case "submit":
                this.flushMouseEvent();
                break;
            case "change":
                if (event.targetType == "select") {
                    // there will be another click action first, so we create this and save it
                    this.flushSelectAction();
                    this.selectAction = this.createSelectOption(event)
                }
                break; // don't record file upload action
                if (event.targetType == "input" && event.targetInfo.inputType == "file") {
                    // was this for an input(type: file)?  If so, create a FileUpload action.
                    this.addAction(this.createFileAction(event));
                    getLogger().debug("Create a File action.");
                }
                break;
            case "DOMTitleChanged":
                this.flushMouseEvent();
                if (domEvent.target.defaultView && domEvent.target.defaultView.frameElement) {
                    return;
                }

                var name = domEvent.target.title.replace(/[\'\"]/g,'');

                if (name && LocatorGenerator.trim(name).length > 0 && name != "about:blank") {
                    if (this.transaction.steps.length > 1 && step.actions.length < 1) {
                        var title = (this.transaction.steps.length - 1) + ". " + name;
                        this.transaction.steps[this.transaction.steps.length - 2].title = title;
                        this.mainWindow.updateStepTitle(this.transaction.steps.length - 2, title);
                    }

                    var title = this.transaction.steps.length + ". " + name;
                    step.title = title;
                    this.mainWindow.updateStepTitle(this.transaction.steps.length - 1, title);
                    if (!step.url || step.url == "about:blank" && domEvent.target) {
                        var url = LocatorGenerator.getDocument(domEvent.target).URL;
                        if (url != "about:blank") {
                            step.url = url;
                        }
                    }
                }
                break;
            case "compositionstart":
                this.isComposition = true;
                this.isIME = true;
                break;
            case "compositionend":
                this.isComposition = false;
                break;
            case "scroll":
                break;
            default:
                //TODOMC log
                break;
        }
    } catch(ex) {
        logException(ex,"processDomEvent");
    }
}

RecorderActionStore.prototype.flushSelectAction = function() {
    if (this.selectAction) {
        getLogger().debug("Create a select action in flushSelectAction.");
        this.addAction(this.selectAction);
        this.selectAction = null;
    }
}

RecorderActionStore.prototype.flushMouseEvent = function() {
    if (this.needFlushMouseEvent) {
        getLogger().debug("Create a click action in flushMouseEvent.");
        this.addAction(this.createClick(this.lastMouseEvent));
        this.needFlushMouseEvent = false;
        this.flushSelectAction();
    }
}

RecorderActionStore.prototype.handleKeyPress = function(step, event) {
    var keyCode;
    if (this.lastKeyDown) {
        keyCode = this.lastKeyDown.eventOptions.keyCode;
    }
    var charCode = event.eventOptions.charCode;
    var action = this.createKeytIn(event);
    if (this.lastKeyDown) {
        action.keyCodes[0] = this.lastKeyDown.eventOptions.keyCode;
        // always use the keydown target, not the keypress target.
        action.window = this.lastKeyDown.windowPath;
        action.locators = this.lastKeyDown.targetLocators;
        // clear the lastKeyDown
        this.lastKeyDown = null;
    }
    this.lookforNetwork = true;
    var modifyLookNetwork = false;
    var lastAction = step.actions[step.actions.length - 1];
    if (lastAction && lastAction.type=="wait") {
        lastAction = step.actions[step.actions.length-2];
        modifyLookNetwork = true;
    }
    var changes = {};
    if (this.addKeytIn(lastAction, action, changes)) {
        getLogger().debug("Create a keytin action.");
        this.addAction(action);
    } else {
        action = lastAction;
        if (modifyLookNetwork) this.lookforNetwork = false;
    }
    var domTarget = event.targetInfo.domTarget;
    var self = this;
    if (keyCode != 9 && keyCode != 13) {
        var prop = "textValue";
        this.keytinUpdate = function() {
            self.keytinUpdate = null;
            var newValue = event.retrieveTextValue(domTarget);
            if (newValue != undefined) {
                var origValue = action[prop];
                action[prop] = newValue;
                var changes = {};
                changes[prop] = newValue;
            }
        };
    }
}

// Returns true if we need to append the newAction; false if we do not need to and
// have appended it to the lastAction.
RecorderActionStore.prototype.addKeytIn = function(lastAction, newAction, changes) {
    if (newAction.type == "keytIn") {
        if (!lastAction) return true;
        if (lastAction.type != "keytIn") return true;
        if (newAction.type != "keytIn") return true;
        if (!this.locatorsEqual(lastAction.locators[0],newAction.locators[0])) return true;
        if (lastAction.modifiers.length != newAction.modifiers.length) return true;
        for (var i=0; i<lastAction.modifiers.length; i++) {
            if (lastAction.modifiers[i] != newAction.modifiers[i]) return true;
        }

        lastAction.keyCodes.push(newAction.keyCodes[0]);
        lastAction.charCodes.push(newAction.charCodes[0]);
        lastAction.selectionStart.push(newAction.selectionStart[0]);
        lastAction.selectionEnd.push(newAction.selectionEnd[0]);
        lastAction.textValue = newAction.textValue;

        changes.keyCodes = lastAction.keyCodes;
        changes.charCodes = lastAction.charCodes;
        changes.selectionStart = lastAction.selectionStart;
        changes.selectionEnd = lastAction.selectionEnd;
        changes.textValue = lastAction.textValue;
        return false;
    } else {
        if (!lastAction) return true;
        if (lastAction.type != "setInputValue") return true;
        if (newAction.type != "setInputValue") return true;
        if (!this.locatorsEqual(lastAction.locators[0], newAction.locators[0])) return true;
        lastAction.value = newAction.value;
        changes.value = lastAction.value;
        return false;
    }
}

RecorderActionStore.prototype.generateScrolls = function() {
    return getRecordScrolls();
}

RecorderActionStore.prototype.handleNetworkEvent = function(timestamp, networkEvent) {
    if(networkEvent) {
        this.processNetworkEvent(timestamp, networkEvent);
    }
}

RecorderActionStore.prototype.processNetworkEvent = function(timestamp, event) {
    if(!event) {
        //TODOMC log this
        return;
    }

    // Save the event for debugging purposes.
    this.events.push(event);
    var step = this.getCurrentStep();

    switch(event.type) {
        case "navigate":
            this.flushMouseEvent();
            getLogger().debug("Create a navigate action. URL: " + event.uri);
            this.addAction(this.createNavigate(event.uri));

            // don't insert waitForNetworks on statechange:start.
            this.lookforNetwork = false;
            break;
        case "locationchange":
            // the URL is changing, which indicated navigation, so we need
            // to insert a wait for page_complete and a new step.
            // make sure that this "new" URL is not just a # on the old URL.
            
            if (step.actions.length > 0 && event.causedRequest) {
                this.flushMouseEvent();
                // don't pay attention if there aren't any actions -- no use
                // for steps with only network waits.
                var lastEvent = step.actions[step.actions.length - 1];

                if (lastEvent.type == "wait" && (lastEvent.style == "network" || lastEvent.style == "page_complete")) {
                    this.deleteAction();
                }
                // TODOMC ask Geoff about this
                getLogger().debug("Create a wait action for page complete in locationchange.");
                this.addAction(this.createWaitFor("page_complete"));

                step.url = event.uri;

                if (1 == this.script.length) {
                    var nameStart = event.uri.indexOf("://") + 3;
                    var name = event.uri.substr(nameStart);
                    name = name.substr(0, name.indexOf("/"));
                    this.transaction.name = name;
                }

                this.addStep();
                // we are not interested in adding waitForNetwork anymore.
                this.lookforNetwork = false;
            }
            break;
        case "statechange":
            if(event.isStart && this.lookforNetwork && event.uri.indexOf('about:') != 0) {
                this.flushMouseEvent();
                getLogger().debug("Create a wait action for network in statechange.");
                this.addAction(this.createWaitFor("network"), true);
                this.lookforNetwork = false;
            }
            break;
        case "statuschange":
            if(this.lookforNetwork && event.uri.indexOf('about:') != 0) {
                this.flushMouseEvent();
                getLogger().debug("Create a wait action for network in statuschange.");
                this.addAction(this.createWaitFor("network"), true);
                this.lookforNetwork = false;
            }
            break;
        case "securitychange":
            // set security on the step
            // get previous step
            step = this.getPreviousStep();
            if (!step) {
                step = this.getCurrentStep();
            }
            if (step && event.credentials) {
                step.httpUser = event.credentials.user;
                step.httpPassword = event.credentials.password;
            }
            break;
        default:
            break;
    }
}

RecorderActionStore.prototype.finish = function(timestamp) {
    this.flushMouseEvent();

    if (this.getCurrentStep().actions.length == 0) {
        this.script.pop();
        this.mainWindow.deleteStep();
    }
    return this.transaction;
}

RecorderActionStore.prototype.getCurrentStep = function() {
    if(this.currentStep) {
        return this.currentStep;
    } else {
        getLogger().debug("Create a transaction.")
        this.transaction = {
            "id": 110,
            "name": "myscript",
            "enableHeader" : false,
            "state": 0,
            "troe": true,
            "steps": this.script
        };
        this.addStep();
    }

    return this.currentStep;
}

RecorderActionStore.prototype.getPreviousStep = function() {
    if (this.script.length > 1) {
        return this.script[this.script.length - 2];
    }
    return null;
}

RecorderActionStore.prototype.addStep = function() {
    // Create it
    var newStep = this.createStep(this.script.length);

    // Add it
    this.script.push(newStep);
    this.mainWindow.addStep();

    // Make it current
    this.currentStep = newStep;
}

RecorderActionStore.prototype.createStep = function(stepNum) {
    getLogger().debug("Create step" + stepNum + ".");
    return {
        "id" : stepNum,
        "title" : "about:blank",
        "url" : "about:blank",
        "actions":[]
    };
}

RecorderActionStore.prototype.locatorsEqual = function(locators1, locators2) {
    //getLogger().debug("----------compare \"" + JSON.stringify(locators1) + "\" and \"" + JSON.stringify(locators2) + "\"");
    return JSON.stringify(locators1) == JSON.stringify(locators2);
}

RecorderActionStore.prototype.createNavigate = function(uri) {
    return {
        "type": "navigate",
        "url": uri
    };
}

RecorderActionStore.prototype.createWaitFor = function(style, time = 0) {
    return {
        "type": "wait",
        "style": style,
        "timeout": time? time : 60000 };
}

RecorderActionStore.prototype.createClick = function(event, bDouble = false) {
    var click = {
        "type": "click",
        "window" : event.windowPath,
        "locators" : event.targetLocators,
        "doubleclick" : bDouble
    };

    if (event.targetInfo.clientWidth || event.targetInfo.clientHeight){
        click.clientWidth = event.targetInfo.clientWidth;
        click.clientHeight = event.targetInfo.clientHeight;
    }

    return click;
}

RecorderActionStore.prototype.createSelectOption = function(event) {
    return {
        "type": "selectOption",
        "window" : event.windowPath,
        "locators" : event.targetLocators,
        "optionIndexes" : event.selectedIndexes,
        "textValues" : event.targetInfo.textValue
    };
}

RecorderActionStore.prototype.createKeytIn = function(event) {
    var modifiers = [];
    if (event.eventOptions.ctrlKey) modifiers.push("CTRL");
    if (event.eventOptions.altKey) modifiers.push("ALT");
    if (event.eventOptions.shiftKey) modifiers.push("SHIFT");
    if (event.eventOptions.metaKey) modifiers.push("META");
    return {
        "type" : "keytIn",
        "window" : event.windowPath,
        "locators" : event.targetLocators,
        "keyCodes": [ event.eventOptions.keyCode ],
        "charCodes": [ event.eventOptions.charCode ],
        "selectionStart": [ event.targetInfo.selectionStart ],
        "selectionEnd": [ event.targetInfo.selectionEnd ],
        "textValue": event.targetInfo.textValue,
        "modifiers": modifiers,
    };
}

RecorderActionStore.prototype.createFileAction = function(event) {
    return {
        "type" : "fileUpload",
        "window" : event.windowPath,
        "locators" : event.targetLocators
    };
}

RecorderActionStore.prototype.addAction = function(action, bPreStep = false) {
    if (('wait' == action.type && !this.breakRecord) || this.actionNum < this.actionMaxNum) {
        var step = this.getCurrentStep();
        //getLogger().debug("bPreStep is " + bPreStep + ", breakRecord is " + this.breakRecord);
        if (bPreStep && step.actions.length == 0 && this.script.length > 1) {
            this.script[this.script.length-2].actions.push(action);
            this.mainWindow.addAction(action.type, true);
        } else {
            step.actions.push(action);
            this.mainWindow.addAction(action.type);
        }
        this.actionNum++;
    } else {
        this.breakRecord = true;
        //getLogger().debug("set breakRecord to true");
        this.mainWindow.breakRecord();
    }
}

RecorderActionStore.prototype.deleteAction = function() {
    this.currentStep.actions.pop();
    this.mainWindow.deleteAction();
    this.actionNum--;
}