var Cu = Components.utils;
Cu.import("resource://oneapmrecorder/ComponentLogger.jsm");
var getLogger = ComponentLoggerFactory.getLoggerFunc("RecDomEvent");
// Holds DOM events we care about but don't want to playback explicitly
function RecDomEvent(timeStamp, domEvent) {
    if(!timeStamp || !domEvent) {
        throw {
            name : "ArgumentException",
            message : "RecDomEvent.createScriptEvent requires valid timeStamp and domEvent arguments"
        }
    }

    this.recType = "dom";
    if (domEvent.target.localName)
        this.targetType = domEvent.target.localName.toLowerCase();
    else
        this.targetType = domEvent.target.nodeName;
    if (this.targetType == "select") {
        this.selectedIndexes = [];
        for (var i=0;i<domEvent.target.options.length;i++) {
            if (domEvent.target.options[i].selected) {
                this.selectedIndexes.push(i);
            }
        }
    }
    this.timeStamp = timeStamp;
    this.eventType = domEvent.type;
    this.windowPath = LocatorGenerator.getFramePath(domEvent.target);
    this.targetLocators = LocatorGenerator.getLocators(domEvent.target);
    if (domEvent.explicitOriginalTarget && domEvent.explicitOriginalTarget != domEvent.target) {
        this.origTarget = LocatorGenerator.getLocators(domEvent.explicitOriginalTarget);
    }
    this.eventOptions = this.createEventOptions(domEvent);
    this.targetInfo = this.createTargetInfo(domEvent);
}

RecDomEvent.prototype.createEventOptions = function(domEvent) {
    if (/^mouse(over|out|down|up|move)|(dbl)?click$/.test(domEvent.type)) {
        return {
            screenX : domEvent.screenX,
            screenY: domEvent.screenY,
            clientX: domEvent.clientX,
            clientY: domEvent.clientY,
            ctrlKey: domEvent.ctrlKey,
            altKey: domEvent.altKey,
            shiftKey: domEvent.shiftKey,
            metaKey: domEvent.metaKey,
            button: domEvent.button
        };
    } else if (/^key(up|down|press)$/.test(domEvent.type)) {
        return {
            ctrlKey: domEvent.ctrlKey,
            altKey: domEvent.altKey,
            shiftKey: domEvent.shiftKey,
            metaKey: domEvent.metaKey,
            keyCode: domEvent.keyCode,
            charCode: (domEvent.type == "keypress") ? domEvent.charCode : null
        };
    } else {
        return {
            ctrlKey: domEvent.ctrlKey,
            altKey: domEvent.altKey,
            shiftKey: domEvent.shiftKey,
            metaKey: domEvent.metaKey
        };
    }
}


RecDomEvent.prototype.createTargetInfo = function(domEvent) {
    if(!domEvent.target) return {};

    var targetInfo = {};
    if (domEvent.target.localName == "input") {
        targetInfo.inputType = domEvent.target.type;
    }

    if (domEvent.type == "keypress") {
        targetInfo.domTarget = domEvent.target;
    }

    if(domEvent.type == "keypress" || domEvent.type == "keyup" ||
        domEvent.type == "keydown" || domEvent.type == "click" ||
        domEvent.type == "input") {
        var target = domEvent.target;
        targetInfo.textValue = this.retrieveTextValue(target);
        if (domEvent.type == "keypress") {
            targetInfo.selectionStart = -1;
            targetInfo.selectionEnd = -1;
            try {
                targetInfo.selectionStart = target.selectionStart;
                targetInfo.selectionEnd = target.selectionEnd;
            } catch(ex) {}
        }
    }
    if (domEvent.type == "change" && domEvent.target.nodeName.toLowerCase() == "select") {
        targetInfo.textValue = [];
        for (var i=0;i<domEvent.target.options.length;i++) {
            if (domEvent.target.options[i].selected) {
                targetInfo.textValue.push(domEvent.target.options[i].textContent);
            }
        }
    }

    if (domEvent.type == "click" || domEvent.type == "mousedown" || domEvent.type == "mousemove") {
      var target = LocatorGenerator.$(domEvent.target);
      if (this.targetType == "embed1")
      {
         // It appears the pageX/Y-offet doesn't work for embeds, not sure why.
           targetInfo.x = domEvent.clientX;
           targetInfo.y = domEvent.clientY;
      }
      else
      {
         // it's not clear why we didn't just use clientX/clientY for non-embeds instead of the pageX/Y - offset.left/top.
           targetInfo.x = domEvent.pageX - parseInt(target.offset().left);
           targetInfo.y = domEvent.pageY - parseInt(target.offset().top);
      }

      // capture screen dimensions for flash interactions
      if (this.targetType == "embed" || this.targetType == "object"){
         targetInfo.clientWidth = domEvent.currentTarget.clientWidth;
         targetInfo.clientHeight = domEvent.currentTarget.clientHeight;
      }
    }

    if(domEvent.type == "scroll" && domEvent.target.defaultView){
        targetInfo.pageXOffset = domEvent.target.defaultView.pageXOffset;
        targetInfo.pageYOffset = domEvent.target.defaultView.pageYOffset;
    }
    return targetInfo;
}

RecDomEvent.prototype.retrieveTextValue = function(domElement) {
    var currentValue = null;

    if(!domElement || !domElement.tagName) return;
    //log("RecDomEvent " + domElement.outerHTML);
    var tagName = domElement.tagName.toLowerCase();
    switch(tagName) {
        case "input":
            if(domElement.type) {
                var eventType = domElement.type.toLowerCase();

                switch(eventType) {
                    case "text":
                    case "password":
                    case "search":
                    case "tel":
                    case "url":
                    case "email":
                    case "datetime":
                    case "date":
                    case "month":
                    case "week":
                    case "time":
                    case "datetime-local":
                    case "number":
                    case "range":
                    case "color":
                    case "hidden":
                        if(domElement.value) {
                            currentValue = domElement.value;
                        }
                        break;
                    default:
                        break;
                }
            }
            break;
        case "textarea":
            if(domElement.value) {
                currentValue = domElement.value;
            }
            break;
        default:
            break;
    }

    return currentValue
}


var LocatorGenerator = {

    init: function(jqueryFunc, tabLocatorFunc) {
        if(!jqueryFunc || !tabLocatorFunc) {
            throw { error: "ArgumentException", message : "A valid jqueryFunc and tabLocatorFunc are required" };
        }

        this.$ = jqueryFunc;
        this.locateTab = tabLocatorFunc;
    },

    trim: function(str) {
        var str = str.replace(/^\s\s*/, ''),
            ws = /\s/,
            i = str.length;
        while (ws.test(str.charAt(--i)));
        return str.slice(0, i + 1);
    },

    getLocators: function(element) {
        var order = [
                ['dom', LocatorGenerator.getElementIdDom],
                ['css', LocatorGenerator.getElementIdCss],
                ['dom', LocatorGenerator.getElementFormDom],
                ['css', LocatorGenerator.getElementFormCss],
                ['dom', LocatorGenerator.getElementFormInputDom],
                ['css', LocatorGenerator.getElementFormInputCss],
                ['css', LocatorGenerator.getElementLinkCss],
                ['css', LocatorGenerator.getElementContainsCss],
                ['css', LocatorGenerator.getElementImageCssAlt],
                ['css', LocatorGenerator.getElementImageCssSrc],
                ['css', LocatorGenerator.getElementAreaCss],
                ['css', LocatorGenerator.getElementClassCss],
                ['css', LocatorGenerator.getElementPathCss]
        ];
        var locs = [];
        for (var i=0; i<order.length; i++) {
                try {
                        var l = order[i][1](element);
                        if (l)
                            locs.push([order[i][0], l]);
                } catch(ex) {
                    getLogger().debug("Error generating locator "+i+": "+ex.message);
                }
        }
        return locs;
    },

    getDocument: function(element) {
        if (element.nodeType == 9) return element;
        if (element.nodeType != 1) return element.document;
        return element.ownerDocument;
    },

    getFramePath: function(element) {
        var doc = LocatorGenerator.getDocument(element);
        return LocatorGenerator.computeFramePath(doc.defaultView);
    },

    createAndCheckCSS: function (element, selector, equalityFn) {
        if (!equalityFn) {
            equalityFn = function(a, b) {
                return a == b;
            }
        }

        var selection = LocatorGenerator.$(selector, LocatorGenerator.getDocument(element));
        if (selection.length == 0) return null;
        if (selection.length > 1) {
            for (var i=0; i < selection.length; i++) {
                if (selection.get(i) == element) {
                    if (i == 0) break;
                    selector+=":eq("+i+")";
                    break;
                }
            }
        }
        if (equalityFn(element, LocatorGenerator.$(selector, LocatorGenerator.getDocument(element)).get(0)))
            return selector;
        return null;
    },

    getElementClassCss: function getElementClassCss(element) {
        if (element.nodeType == 1 && element.hasAttribute('class')) {
            var classes = element.getAttribute('class').split(' ');
            for (var i=0; i<classes.length; i++) {
                if (classes[i].length > 0) {
                    var loc = "." + classes[i];
                    var el = LocatorGenerator.createAndCheckCSS(element, loc);
                    if (el) return el;
                }
            }
        }
        return null;
    },

    getElementImageCssAlt: function getElementImageCssAlt(element) {
        if (element.nodeType == 1 && element.localName.toLowerCase() == "img") {
            if (element.hasAttribute("alt") && LocatorGenerator.trim(element.alt).length > 0) {
              if (element.alt.match(/[\'\"]+/) != null){
                   return null;
               }
               var loc = 'img[alt="'+ element.alt +'"]';

            return LocatorGenerator.createAndCheckCSS(element, loc);
            }
        }
        return null;
    },

    getElementImageCssSrc: function getElementImageCssSrc(element) {
        if (element.nodeType == 1 && element.localName.toLowerCase() == "img") {
            return LocatorGenerator.generateLink(element, 'src');
        }
        return null;
    },

    getElementAreaCss: function getElementAreaCss(element) {
        if (element.nodeType == 1 && element.localName.toLowerCase() == "area") {
            return LocatorGenerator.generateLink(element, 'href');
        }
        return null;
    },

    getElementFormCss: function getElementFormCss(element) {
        if (element.nodeType == 1 && element.localName.toLowerCase() == "form") {
            if (element.hasAttribute('name')) {
                return LocatorGenerator.createAndCheckCSS(element, 'form[name="'+element.getAttribute('name')+'"]');
            }
        }
        return null;
    },

    getElementFormInputCss: function getElementFormInputCss(element) {
        if (element.nodeType == 1 &&  ["input","textarea","select","button"].indexOf(element.localName.toLowerCase()) != -1) {
            var inputElement = element;
            var selector = element.localName.toLowerCase();
            if (inputElement.hasAttribute('type')) {
                selector += '[type="'+inputElement.type.toLowerCase()+'"]';
                if (inputElement.type == "image") {

                    if (inputElement.hasAttribute('alt')) {
                        if (inputElement.alt.match(/[\'\"]+/) != null){
                           //log('found quote');
                           return null;
                        }
                        selector += '[alt="'+inputElement.alt +'"]';
                    } else if (inputElement.hasAttribute('src')) {
                        selector += '[src$="'+inputElement.getAttribute('src').replace(/'/g,"\'")+'"]';
                    }
                }
            }
            if (inputElement.hasAttribute('name')) {
                selector += '[name="'+inputElement.name.replace(/'/g,"\'")+'"]';
            }
            return LocatorGenerator.createAndCheckCSS(inputElement, selector);
        } else if (element.nodeType == 1 && element.localName.toLowerCase() == "option") {
            var optionEl = element;
            var selectEl = element.parentNode;
            while (selectEl.nodeType ==1 && selectEl.localName.toLowerCase() != "select") {
                selectEl = selectEl.parentNode;
            }
            if (selectEl.nodeType != 1 || selectEl.localName.toLowerCase() != "select") return null;
            var selectPath = LocatorGenerator.getElementFormInputCss(selectEl);
            if (selectPath == null) return null;
            selectPath += " option";
            // try contains:
            var text = optionEl.textContents;
            var textSelect = LocatorGenerator.createAndCheckCSS(optionEl, selectPath + ':contains("'+text+'")');
            if (textSelect)
                return textSelect;
            return LocatorGenerator.createAndCheckCSS(optionEl, selectPath);
        }
        return null;
    },

    jq: function (myid) {
        return '#'+myid.replace(/:/g,"\\:").replace(/\./g,"\\.").replace(/\[/g, "\\[").replace(/\]/g,"\\]").replace(/\//g,"\\/");
    },

    getElementIdCss: function getElementIdCss(element) {
        if (element.nodeType == 1 && element.hasAttribute('id')) {
            return LocatorGenerator.createAndCheckCSS(element, LocatorGenerator.jq(element.id));
        }
        return null;
    },

    getElementIdDom: function getElementIdDom(element) {
        if (element.nodeType == 1 && element.hasAttribute('id')) {
            if (LocatorGenerator.getDocument(element).getElementById(element.id) == element)
                return 'document.getElementById("'+element.id+'")';
        }
        return null;
    },

    getElementPathCss: function getElementPathCss(element) {
        var paths = [];
        var origElement = element;
        for (; element && element.nodeType == 1; element = element.parentNode) {
            // escape for IDs
            if (element.id && element != origElement) {
                var id = LocatorGenerator.jq(element.id);
                paths.splice(0,0,id);
                break;
            }
            // escape for name
            if (element != origElement &&
                element.hasAttribute('name')) {
                paths.splice(0,0,element.localName.toLowerCase() + '[name="'+element.name.replace(/'/g,"\'")+'"]');
                break;
            }
            // fallback to index + parent
            var index = 0;
            for (var sibling = element.previousSibling; sibling; sibling = sibling.previousSibling) {
                if (sibling.localName == element.localName)
                    index++;
            }
            var tagName = element.localName.toLowerCase();
            var pathIndex = (index ? ":eq(" + (index) + ")" : "" );
            paths.splice(0,0,tagName + pathIndex);
        }

        return paths.length ?
            paths.join(" ") :
            null;
    },

    findByIdNameIndexInCollectionDom: function(element, collection, prefix, escapeViaId) {
        if (element.nodeType == 1 && element.hasAttribute("id")) {
            if (collection.item(element.id) == element)
                return prefix + '["' + element.id + '"]';
            else if (escapeViaId)
                return "document.getElementById('"+element.id+"')";
        }

        if (element.nodeType == 1 && element.hasAttribute("name") && collection.item(element.name) == element) {
            return prefix + '["' + element.name + '"]';
        }
        // fallback to indexing
        var index = -1;
        for (var i = 0; i < collection.length; i++) {
            if (collection.item(i) == element) {
                index = i;
                break;
            }
        }
        if (index > -1) {
            return prefix + '[' + index +']';
        }
        return null;
    },

    getElementFormDom: function getElementFormDom(form) {
        if (!form || form.nodeType != 1 || form.localName.toLowerCase() != "form") return null;
        var path = LocatorGenerator.findByIdNameIndexInCollectionDom(form, form.ownerDocument.forms, "document.forms", false);
        if (!path) {
            // no way to find the form apparently, so fail for DOM.
            return null;
        }
        return path;
    },

    getElementFormInputDom: function getElementFormInputDom(element) {
        if (element.nodeType == 1 && ["input","textarea","select","button"].indexOf(element.localName.toLowerCase()) != -1) {
            var inputElement = element;
            if (inputElement.localName.toLowerCase() == "input" && inputElement.type == "image") {
                // input images can only be accessed via children, not elements.
                return null;
            }
            var form = element.form;
            var path = LocatorGenerator.getElementFormDom(form);
            if (!path) {
                // no way to find the form apparently, so fail for DOM.
                return null;
            }
            // find element in elements array; .elements is not needed, direct forms['a']['b'] works.
            var elPath = LocatorGenerator.findByIdNameIndexInCollectionDom(inputElement, form.elements, "", false);
            if (!elPath) {
                return null;
            }
            return path + elPath;
        } else if (element.nodeType == 1 && element.localName.toLowerCase() == "option") {
            // find the select it's attached to
            var optionEl = element;
            var selectEl = element.parentNode;
            while (selectEl.nodeType == 1 && selectEl.localName.toLowerCase() != "select") selectEl = selectEl.parentNode;
            if (selectEl.nodeType != 1 || selectEl.localName.toLowerCase() != "select") return null;
            var selectPath = LocatorGenerator.getElementFormInputDom(selectEl);
            if (selectPath) {
                // find option index
                var optionPath = LocatorGenerator.findByIdNameIndexInCollectionDom(optionEl, selectEl.options, ".options", false);
                if (optionPath)
                    return selectPath+optionPath;
            }
        }
        return null;
    },


    findTextContent: function(element) {
        // find all text content recursively
        if (element.nodeType == 3)
            return element.textContent;
        else {
            if (element.nodeType == 1) {
                if (element.localName.toLowerCase() == "script") return "";
                var text = "";
                for (var i=0; i<element.childNodes.length; i++) {
                    text += LocatorGenerator.findTextContent(element.childNodes[i]) + "\n";
                }
                return text;
            } else {
                return "";
            }
        }
    },

    findLongestTextContent: function(element) {
        var textContent = LocatorGenerator.findTextContent(element);  //element.textContent;
        // split on tabs and newlines
        // split on parens because sizzle/jquery blows up if contains includes (,)
        textContent = LocatorGenerator.trim(textContent);
        textContent = textContent.replace(/\"/g,'\\"');
        textContent = textContent.split('\n');
        var longestContent = "";
        while (textContent.length > 0) {
            var splitContent = textContent.pop().split('(');
            while(splitContent.length > 0) {
                var moreSplit = splitContent.pop().split(')');
                for (var i=0;i<moreSplit.length;i++) {
                    var str = LocatorGenerator.trim(moreSplit[i]);
                    if (str.length > longestContent.length) longestContent = str;
                }
            }
        }
        return longestContent;
    },

    getElementLinkCss: function getElementLinkCss(element) {
        // search for an <a> parent
        element = LocatorGenerator.findParentNode(element, ['a']);
        var checkResultA = function(orig, path) {
            var result = LocatorGenerator.$(path, LocatorGenerator.getDocument(orig)).get(0);
            return (result &&
                    result.textContent == orig.textContent &&
                    result.getAttribute('href') == orig.getAttribute('href') &&
                    result.title == orig.title);
        }
        var findIndexOfA = function(orig, path) {
            var result = LocatorGenerator.$(path, LocatorGenerator.getDocument(orig));
            if (result.length > 0) {
                return LocatorGenerator.createAndCheckCSS(orig, path, checkResultA);
            }
            return null;
        };

        if (element) {
            var longestContent = LocatorGenerator.findLongestTextContent(element);
            if (longestContent && longestContent.length > 0) {
                var path = 'a:contains("'+longestContent+'")';
                var path = findIndexOfA(element, path);
                if (path) return path;
            }
            var hrefPath = LocatorGenerator.generateLink(element, 'href', checkResultA);
            if (hrefPath) {
                return hrefPath;
            }
            if (element.hasAttribute('title') && element.title && LocatorGenerator.trim(element.title).length > 0) {
                path = 'a[title="'+element.title.replace(/'/g,"\'")+'"]';
                var path = findIndexOfA(element, path);
                if (path) return path;
            }
        }
        return null;
    },

    generateLink: function generateLink(element, attr, equalityFn) {
        if (!element.hasAttribute(attr)) return null;
        var value = element.getAttribute(attr);
        if (!value) return null;
        var valueTrimmed = LocatorGenerator.trim(value);
        if (valueTrimmed.length < 1) return null;
        if (valueTrimmed.toLowerCase().indexOf('javascript') == 0) return null;
        if (valueTrimmed == "#") return null;
        if (valueTrimmed.indexOf("'") != -1) return null;
        if (valueTrimmed.indexOf('"') != -1) return null;
        // use $= on links, so we don't worry about resolved vs. specified differences
        // in IE/FF.
        var path = element.localName.toLowerCase() + '[' + attr + '$="' + value +'"]';
        var result = LocatorGenerator.$(path, LocatorGenerator.getDocument(element));
        if (result.length > 0) {
            return LocatorGenerator.createAndCheckCSS(element, path, equalityFn);
        }
        return null;
    },

    findParentNode: function(element, parentTypes, depth) {
        if (depth == undefined) depth = 100;
        var currentDepth = 0;
        while (element && element.nodeType == 1 && currentDepth < depth &&
               parentTypes.indexOf(element.localName.toLowerCase()) == -1) {
            element = element.parentNode;
            currentDepth++;
        }
        if (element && element.nodeType == 1 && currentDepth < depth &&
            parentTypes.indexOf(element.localName.toLowerCase() != -1))
            return element;
        return null;
    },


    getElementContainsCss: function getElementContainsCss(element) {
        var origElement = element;
        // look for an immediate parent that is div, span, td, button, input
        var element = LocatorGenerator.findParentNode(element, ['div', 'span', 'td', 'button', 'input'], 2);

        if (element) {
            // check to see if they have a valid contains
            var longestContent = LocatorGenerator.findLongestTextContent(element);
            if (longestContent && longestContent.length > 0) {
                var path = element.nodeName.toLowerCase()+':contains("'+longestContent+'")';
                var result = LocatorGenerator.$(path, LocatorGenerator.getDocument(origElement));
                if (result.length > 0) {
                    path = LocatorGenerator.createAndCheckCSS(element, path);
                    if (path) return path;
                }
            }
        }
        return null;
    },

    // args: Window
    computeFramePath: function(targetWindow) {
        if (targetWindow.framePath) return targetWindow.framePath;

        if (targetWindow.parent == targetWindow) {
            targetWindow.framePath = LocatorGenerator.locateTab(targetWindow);
            return targetWindow.framePath;
        }

        targetWindow.framePath = LocatorGenerator.computeFramePath(targetWindow.parent);
        targetWindow.framePath += ".frames[";
        for (var i=0; i<targetWindow.parent.frames.length;i++) {
            if (targetWindow.parent.frames[i] == targetWindow) {
                targetWindow.framePath += i +"]";
            }
        }

        return targetWindow.framePath;
    }
};

