var Cc = Components.classes;
var Ci = Components.interfaces;
var Cu = Components.utils;
var Cr = Components.results;
Cu.import("resource://oneapmrecorder/ComponentLogger.jsm");
XULNS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
HTMLNS = "http://www.w3.org/1999/xhtml";
var getLogger = ComponentLoggerFactory.getLoggerFunc("Recorder");
var isRecording = false;
var didInitRecorder = false;
var didSave = false;
var transAction;
var navigateDetail;
var waitDetail;
var clickDetail;
var typeDetail;
var selectDetail;
var actionMaxNum;
var isPlayBacking = false;

function logException(ex, name){
    var msg = ex + " filename: " + ex.fileName ;
    msg += " lineNumber: " + ex.lineNumber ;
    getLogger().error("caught exception in " + name + " : " + msg);
}

function initControl() {
    try {
        getLogger().debug("init element controls");
        RecButton = document.getElementById("RecordBtn");
        SaveButton = document.getElementById("SaveBtn");
        OpenButton = document.getElementById("OpenBtn");
        PlayButton = document.getElementById("PlayBtn");
        stepQueue = document.getElementById("ShowBox");
        recycleBin = document.getElementById("recycle-bin");
        mainBox = document.getElementById("main-box");
        var configUrl = "resource://oneapmrecorder/config.ini";
        var configFile = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);
        configFile.open("GET", configUrl, false);
        configFile.send(null);
        var configSetting = parseINIString(configFile.responseText);
        actionMaxNum = configSetting.ActionMaxNum.value;
        getLogger().debug("actionMaxNum is " + actionMaxNum)
    } catch (ex) {
        logException(ex, "initControl");
    }
}

function parseINIString(data){
    var regex = {
        section: /^\s*\[\s*([^\]]*)\s*\]\s*$/,
        param: /^\s*([\w\.\-\_]+)\s*=\s*(.*?)\s*$/,
        comment: /^\s*;.*$/
    };
    var value = {};
    var lines = data.split(/\r\n|\r|\n/);
    var section = null;
    lines.forEach(function(line){
        if(regex.comment.test(line)){
            return;
        }else if(regex.param.test(line)){
            var match = line.match(regex.param);
            if(section){
                value[section][match[1]] = match[2];
            }else{
                value[match[1]] = match[2];
            }
        }else if(regex.section.test(line)){
            var match = line.match(regex.section);
            value[match[1]] = {};
            section = match[1];
        }else if(line.length == 0 && section){
            section = null;
        };
    });
    return value;
}

function addAction(actionType, bPreStep = false) {
    try {
        var newButton = document.createElementNS(XULNS, "button");
        newButton.setAttribute("id", actionType);
        newButton.setAttribute("class", "action");
        newButton.setAttribute("orient", "vertical");
        switch (actionType) {
            case "navigate":
                newButton.setAttribute("label", "导航");
                newButton.setAttribute("image", "chrome://oneapmrecorder/skin/icons/navigate.png");
                break;
            case "wait":
                newButton.setAttribute("label", "等待");
                newButton.setAttribute("image", "chrome://oneapmrecorder/skin/icons/wait.png");
                break;
            case "click":
                newButton.setAttribute("label", "点击");
                newButton.setAttribute("image", "chrome://oneapmrecorder/skin/icons/click.png");
                break;
            case "keytIn":
                newButton.setAttribute("label", "键入");
                newButton.setAttribute("image", "chrome://oneapmrecorder/skin/icons/type.png");
                break;
            case "fileUpload":
                newButton.setAttribute("label", "上传文件");
                newButton.setAttribute("image", "chrome://oneapmrecorder/skin/icons/upload.png");
                break;
            case "selectOption":
                newButton.setAttribute("label", "选择");
                newButton.setAttribute("image", "chrome://oneapmrecorder/skin/icons/select.png");
                break;
        }
        newButton.setAttribute("disabled",true);

        if (bPreStep) {
            stepQueue.childNodes[stepQueue.childNodes.length - 2].firstChild.lastChild.appendChild(newButton);
        } else {
            stepQueue.lastChild.firstChild.lastChild.appendChild(newButton);
        }
    } catch (ex) {
        logException(ex,"addAction");
    }
}

function deleteAction() {
    if (stepQueue.lastChild && stepQueue.lastChild.firstChild && stepQueue.lastChild.firstChild.lastChild) {
        stepQueue.lastChild.firstChild.lastChild.removeChild(stepQueue.lastChild.firstChild.lastChild.lastChild);
    }
}

function enableActionDetail() {
    if (transAction.steps.length != stepQueue.childNodes.length) getLogger().error("Step num is incorrect!");

    for (var stepIndex = 0; stepIndex < stepQueue.childNodes.length; stepIndex++) {
        stepQueue.childNodes[stepIndex].firstChild.firstChild.disabled = false;
        stepQueue.childNodes[stepIndex].firstChild.firstChild.setAttribute("oninput","onModifyStepTitle(\""+ stepIndex + "\", this.value = checkLength(this.value))");
        if (transAction.steps[stepIndex].actions.length != stepQueue.childNodes[stepIndex].firstChild.lastChild.childNodes.length) getLogger().debug("Action num is incorrect!");
        for (var actionIndex = 0; actionIndex < stepQueue.childNodes[stepIndex].firstChild.lastChild.childNodes.length; actionIndex++) {
            stepQueue.childNodes[stepIndex].firstChild.lastChild.childNodes[actionIndex].disabled = false;
            if (transAction.steps.length > stepIndex && transAction.steps[stepIndex].actions.length > actionIndex &&
                transAction.steps[stepIndex].actions[actionIndex].type == stepQueue.childNodes[stepIndex].firstChild.lastChild.childNodes[actionIndex].id) {
                stepQueue.childNodes[stepIndex].firstChild.lastChild.childNodes[actionIndex].setAttribute("oncommand","onShowActionDetail(\""+ stepIndex + "\",\"" + actionIndex + "\")");
            } else {
                getLogger().error("Action type is incorrect!");
            }
        }
    }
}

function setActionEnable(bEnable) {
    for (var stepIndex = 0; stepIndex < stepQueue.childNodes.length; stepIndex++) {
        stepQueue.childNodes[stepIndex].firstChild.firstChild.disabled = !bEnable;
        for (var actionIndex = 0; actionIndex < stepQueue.childNodes[stepIndex].firstChild.lastChild.childNodes.length; actionIndex++) {
            stepQueue.childNodes[stepIndex].firstChild.lastChild.childNodes[actionIndex].disabled = !bEnable;
        }
    }
}

function checkLength(value) {
    var n = 0;
    for (i = 0; i < value.length; i++) {
       var leg = value.charCodeAt(i);//ASCII码
        if (leg > 255) {//大于255的都是中文
            n += 2;//如果是中文就是2个字节
        } else {
            n += 1;//英文
        }
        if (n > 50) {
            alert("标题不能超过50个字符！（一个汉字为两个字符）");
            return value.substr(0,i);
        }
    }
    return value;
}

function onModifyStepTitle(stepIndex, value) {
    transAction.steps[stepIndex].title = value;
}

function onShowActionDetail(stepIndex, actionIndex) {
    try {
        //getLogger().debug("Show detail of step " + stepIndex + " action " + actionIndex + ".");
        var action = transAction.steps[stepIndex].actions[actionIndex];
        if (mainBox.lastChild) recycleBin.appendChild(mainBox.lastChild);

        switch (action.type) {
            case "navigate":
                showNavigateDetail(action);
                break;
            case "wait":
                showWaitDetail(action);
                break;
            case "click":
                showClickDetail(action);
                break;
            case "keytIn":
                showTypeDetail(action);
                break;
            case "selectOption":
                showSelectDetail(action);
                break;
            default:
                break;
        }
    } catch (ex) {
        logException(ex,"onShowActionDetail");
    }
}

function showNavigateDetail(action) {
    if (!navigateDetail) {
        var urlLabel = document.createElementNS(XULNS, "label");
        urlLabel.setAttribute("class", "detail");
        urlLabel.setAttribute("value","网址：");
        urlLabel.setAttribute("width", 100);
        var urlInput = document.createElementNS(XULNS, "textbox");
        urlInput.setAttribute("id", "urlText");
        urlInput.setAttribute("flex", "1");
        urlInput.setAttribute("minWidth", 200);
        urlInput.setAttribute("disabled", true);
        var urlBox = document.createElementNS(XULNS, "hbox");
        urlBox.setAttribute("hflex", "1");
        urlBox.appendChild(urlLabel);
        urlBox.appendChild(urlInput);
        urlBox.setAttribute("align", "center");

        var windowLabel = document.createElementNS(XULNS, "label");
        windowLabel.setAttribute("class", "detail");
        windowLabel.setAttribute("value", "窗口：");
        windowLabel.setAttribute("width", 100);
        var windowInput = document.createElementNS(XULNS, "textbox");
        windowInput.setAttribute("width", 200);
        windowInput.setAttribute("value", "top[0]");
        windowInput.setAttribute("disabled", true);
        var windowBox = document.createElementNS(XULNS, "hbox");
        windowBox.appendChild(windowLabel);
        windowBox.appendChild(windowInput);
        windowBox.setAttribute("align", "center");

        navigateDetail = document.createElementNS(XULNS, "vbox");
        navigateDetail.setAttribute("hflex", 1);
        navigateDetail.appendChild(urlBox);
        navigateDetail.appendChild(windowBox);
        recycleBin.appendChild(navigateDetail);
    }

    document.getElementById("urlText").setAttribute("value", action.url);
    mainBox.appendChild(navigateDetail);
}

function showWaitDetail(action) {
    if (!waitDetail) {
        var styleLabel = document.createElementNS(XULNS, "label");
        styleLabel.setAttribute("class", "detail");
        styleLabel.setAttribute("value", "等待类型：");
        styleLabel.setAttribute("width", 100);
        var styleInput = document.createElementNS(XULNS, "textbox");
        styleInput.setAttribute("id", "styleText");
        styleInput.setAttribute("width", 200);
        styleInput.setAttribute("disabled", true);
        var styleBox = document.createElementNS(XULNS, "hbox");
        styleBox.appendChild(styleLabel);
        styleBox.appendChild(styleInput);
        styleBox.setAttribute("align", "center");

        var timeLabel = document.createElementNS(XULNS, "label");
        timeLabel.setAttribute("class", "detail");
        timeLabel.setAttribute("value", "等待时间：");
        timeLabel.setAttribute("width", 100);
        var timeInput = document.createElementNS(XULNS, "textbox");
        timeInput.setAttribute("id", "timeText");
        timeInput.setAttribute("width", 200);
        timeInput.setAttribute("disabled", true);
        var unitLabel = document.createElementNS(XULNS, "label");
        unitLabel.setAttribute("value", "毫秒");
        var timeBox = document.createElementNS(XULNS, "hbox");
        timeBox.appendChild(timeLabel);
        timeBox.appendChild(timeInput);
        timeBox.appendChild(unitLabel);
        timeBox.setAttribute("align", "center");

        waitDetail = document.createElementNS(XULNS, "vbox");
        waitDetail.appendChild(styleBox);
        waitDetail.appendChild(timeBox);
        recycleBin.appendChild(waitDetail);
    }

    document.getElementById("styleText").setAttribute("value", action.style);
    document.getElementById("timeText").setAttribute("value", action.timeout);
    mainBox.appendChild(waitDetail);
}

function showClickDetail(action) {
    if (!clickDetail) {
        var windowLabel = document.createElementNS(XULNS, "label");
        windowLabel.setAttribute("class", "detail");
        windowLabel.setAttribute("value", "窗口：");
        windowLabel.setAttribute("width", 100);
        var windowInput = document.createElementNS(XULNS, "textbox");
        windowInput.setAttribute("id", "clickWin");
        windowInput.setAttribute("width", 200);
        windowInput.setAttribute("disabled", true);
        var windowBox = document.createElementNS(XULNS, "hbox");
        windowBox.appendChild(windowLabel);
        windowBox.appendChild(windowInput);
        windowBox.setAttribute("align", "center");

        var locatorLabel = document.createElementNS(XULNS, "label");
        locatorLabel.setAttribute("class", "detail");
        locatorLabel.setAttribute("value", "定位：");
        locatorLabel.setAttribute("width", 100);
        var locatorSubBox = document.createElementNS(XULNS, "vbox");
        locatorSubBox.setAttribute("id", "clickLocators");
        var locatorBox = document.createElementNS(XULNS, "hbox");
        locatorBox.appendChild(locatorLabel);
        locatorBox.appendChild(locatorSubBox);

        var classLabel = document.createElementNS(XULNS, "label");
        classLabel.setAttribute("class", "detail");
        classLabel.setAttribute("value", "类型：");
        classLabel.setAttribute("width", 100);
        var classInput = document.createElementNS(XULNS, "textbox");
        classInput.setAttribute("id", "classWin");
        classInput.setAttribute("width", 200);
        classInput.setAttribute("disabled", true);
        var classBox = document.createElementNS(XULNS, "hbox");
        classBox.appendChild(classLabel);
        classBox.appendChild(classInput);
        classBox.setAttribute("align", "center");

        clickDetail = document.createElementNS(XULNS, "vbox");
        clickDetail.appendChild(windowBox);
        clickDetail.appendChild(locatorBox);
        clickDetail.appendChild(classBox);
        recycleBin.appendChild(clickDetail);
    }

    document.getElementById("clickWin").setAttribute("value", action.window);
    var locBox = document.getElementById("clickLocators");
    while(locBox.childNodes.length) locBox.removeChild(locBox.lastChild);
    for (var locIndex = 0; locIndex < action.locators.length; locIndex++) {
        var locType = document.createElementNS(XULNS, "textbox");
        locType.setAttribute("value", action.locators[locIndex][0]);
        locType.setAttribute("disabled", true);
        locType.setAttribute("width", 50);
        var locValue = document.createElementNS(XULNS, "textbox");
        locValue.setAttribute("value", action.locators[locIndex][1]);
        locValue.setAttribute("disabled", true);
        locValue.setAttribute("width", 300);
        var locItem = document.createElementNS(XULNS, "hbox");
        locItem.appendChild(locType);
        locItem.appendChild(locValue);
        locBox.appendChild(locItem);
    }
    if (action.doubleclick) {
        document.getElementById("classWin").setAttribute("value", "双击");
    } else {
        document.getElementById("classWin").setAttribute("value", "单击");
    }
    mainBox.appendChild(clickDetail);
}

function showTypeDetail(action) {
    if (!typeDetail) {
        var textLabel = document.createElementNS(XULNS, "label");
        textLabel.setAttribute("class", "detail");
        textLabel.setAttribute("value", "输入的字符：")
        textLabel.setAttribute("width", 100);
        var textInput = document.createElementNS(XULNS, "textbox");
        textInput.setAttribute("id", "typeText");
        textInput.setAttribute("disabled", true);
        textInput.setAttribute("width", 350);
        var textBox = document.createElementNS(XULNS, "hbox");
        textBox.appendChild(textLabel);
        textBox.appendChild(textInput);
        textBox.setAttribute("align", "center");

        var windowLabel = document.createElementNS(XULNS, "label");
        windowLabel.setAttribute("class", "detail");
        windowLabel.setAttribute("value", "窗口：");
        windowLabel.setAttribute("width", 100);
        var windowInput = document.createElementNS(XULNS, "textbox");
        windowInput.setAttribute("id", "typeWin");
        windowInput.setAttribute("width", 200);
        windowInput.setAttribute("disabled", true);
        var windowBox = document.createElementNS(XULNS, "hbox");
        windowBox.appendChild(windowLabel);
        windowBox.appendChild(windowInput);
        windowBox.setAttribute("align", "center");

        var locatorLabel = document.createElementNS(XULNS, "label");
        locatorLabel.setAttribute("class", "detail");
        locatorLabel.setAttribute("value", "定位：");
        locatorLabel.setAttribute("width", 100);
        var locatorSubBox = document.createElementNS(XULNS, "vbox");
        locatorSubBox.setAttribute("id", "typeLocators");
        var locatorBox = document.createElementNS(XULNS, "hbox");
        locatorBox.appendChild(locatorLabel);
        locatorBox.appendChild(locatorSubBox);

        typeDetail = document.createElementNS(XULNS, "vbox");
        typeDetail.appendChild(textBox);
        typeDetail.appendChild(windowBox);
        typeDetail.appendChild(locatorBox);
        recycleBin.appendChild(typeDetail);
    }

    document.getElementById("typeText").setAttribute("value", action.textValue);
    document.getElementById("typeWin").setAttribute("value", action.window);
    var locBox = document.getElementById("typeLocators");
    while(locBox.childNodes.length) locBox.removeChild(locBox.lastChild);
    for (var locIndex = 0; locIndex < action.locators.length; locIndex++) {
        var locType = document.createElementNS(XULNS, "textbox");
        locType.setAttribute("value", action.locators[locIndex][0]);
        locType.setAttribute("disabled", true);
        locType.setAttribute("width", 50);
        var locValue = document.createElementNS(XULNS, "textbox");
        locValue.setAttribute("value", action.locators[locIndex][1]);
        locValue.setAttribute("disabled", true);
        locValue.setAttribute("width", 300);
        var locItem = document.createElementNS(XULNS, "hbox");
        locItem.appendChild(locType);
        locItem.appendChild(locValue);
        locBox.appendChild(locItem);
    }
    mainBox.appendChild(typeDetail);
}

function showSelectDetail(action) {
    if (!selectDetail) {
        var textLabel = document.createElementNS(XULNS, "label");
        textLabel.setAttribute("class", "detail");
        textLabel.setAttribute("value", "选择项：")
        textLabel.setAttribute("width", 100);
        var textInput = document.createElementNS(XULNS, "textbox");
        textInput.setAttribute("id", "selectText");
        textInput.setAttribute("disabled", true);
        textInput.setAttribute("width", 350);
        var textBox = document.createElementNS(XULNS, "hbox");
        textBox.appendChild(textLabel);
        textBox.appendChild(textInput);
        textBox.setAttribute("align", "center");

        var numLabel = document.createElementNS(XULNS, "label");
        numLabel.setAttribute("class", "detail");
        numLabel.setAttribute("value", "选择序号：")
        numLabel.setAttribute("width", 100);
        var numInput = document.createElementNS(XULNS, "textbox");
        numInput.setAttribute("id", "selectNum");
        numInput.setAttribute("disabled", true);
        numInput.setAttribute("width", 350);
        var numBox = document.createElementNS(XULNS, "hbox");
        numBox.appendChild(numLabel);
        numBox.appendChild(numInput);
        numBox.setAttribute("align", "center");

        var windowLabel = document.createElementNS(XULNS, "label");
        windowLabel.setAttribute("class", "detail");
        windowLabel.setAttribute("value", "窗口：");
        windowLabel.setAttribute("width", 100);
        var windowInput = document.createElementNS(XULNS, "textbox");
        windowInput.setAttribute("id", "selectWin");
        windowInput.setAttribute("width", 200);
        windowInput.setAttribute("disabled", true);
        var windowBox = document.createElementNS(XULNS, "hbox");
        windowBox.appendChild(windowLabel);
        windowBox.appendChild(windowInput);
        windowBox.setAttribute("align", "center");

        var locatorLabel = document.createElementNS(XULNS, "label");
        locatorLabel.setAttribute("class", "detail");
        locatorLabel.setAttribute("value", "定位：");
        locatorLabel.setAttribute("width", 100);
        var locatorSubBox = document.createElementNS(XULNS, "vbox");
        locatorSubBox.setAttribute("id", "selectLocators");
        var locatorBox = document.createElementNS(XULNS, "hbox");
        locatorBox.appendChild(locatorLabel);
        locatorBox.appendChild(locatorSubBox);

        selectDetail = document.createElementNS(XULNS, "vbox");
        selectDetail.appendChild(textBox);
        selectDetail.appendChild(numBox);
        selectDetail.appendChild(windowBox);
        selectDetail.appendChild(locatorBox);
        recycleBin.appendChild(selectDetail);
    }

    document.getElementById("selectText").setAttribute("value", action.textValues);
    document.getElementById("selectNum").setAttribute("value", action.optionIndexes);
    document.getElementById("selectWin").setAttribute("value", action.window);
    var locBox = document.getElementById("selectLocators");
    while(locBox.childNodes.length) locBox.removeChild(locBox.lastChild);
    for (var locIndex = 0; locIndex < action.locators.length; locIndex++) {
        var locType = document.createElementNS(XULNS, "textbox");
        locType.setAttribute("value", action.locators[locIndex][0]);
        locType.setAttribute("disabled", true);
        locType.setAttribute("width", 50);
        var locValue = document.createElementNS(XULNS, "textbox");
        locValue.setAttribute("value", action.locators[locIndex][1]);
        locValue.setAttribute("disabled", true);
        locValue.setAttribute("width", 300);
        var locItem = document.createElementNS(XULNS, "hbox");
        locItem.appendChild(locType);
        locItem.appendChild(locValue);
        locBox.appendChild(locItem);
    }
    mainBox.appendChild(selectDetail);
}

function addStep() {
    try {
        var stepLabel = document.createElementNS(XULNS,"textbox");
        stepLabel.setAttribute("class", "stepTitle");
        stepLabel.setAttribute("height",20);
        stepLabel.setAttribute("hflex", 1);
        stepLabel.setAttribute("disabled", true);
        stepLabel.setAttribute("width", 20);
        stepLabel.setAttribute("value", (stepQueue.childNodes.length + 1) + ".");

        var line = document.createElementNS(HTMLNS,"hr");

        var actionsBox = document.createElementNS(XULNS,"hbox");
        actionsBox.setAttribute("height",56);
        actionsBox.setAttribute("align", "center");

        var tempBox = document.createElementNS(XULNS,"vbox");
        tempBox.setAttribute("height",78);
        tempBox.appendChild(stepLabel);
        tempBox.appendChild(line);
        tempBox.appendChild(actionsBox);

        var line2 = document.createElementNS(HTMLNS,"hr");

        var stepBox = document.createElementNS(XULNS,"hbox");
        stepBox.setAttribute("height", 70);
        stepBox.appendChild(tempBox);
        stepBox.appendChild(line2);

        stepQueue.appendChild(stepBox);
    } catch (ex) {
        logException(ex,"addStep");
    }
}

function deleteStep() {
    try {
        if (stepQueue.lastChild) stepQueue.removeChild(stepQueue.lastChild);
    } catch (ex) {
        logException(ex, "deleteStep");
    }
}

function updateStepTitle(stepIndex, title) {
    if (stepIndex < stepQueue.childNodes.length) {
        stepQueue.childNodes[stepIndex].firstChild.firstChild.setAttribute("value", title);
    }
}

function showTransAction() {
    while (stepQueue.childNodes.length > 0) stepQueue.removeChild(stepQueue.firstChild);
    for (var stepIndex = 0; stepIndex < transAction.steps.length; stepIndex++) {
        addStep();
        updateStepTitle(stepIndex, transAction.steps[stepIndex].title);
        for (var actionIndex = 0; actionIndex < transAction.steps[stepIndex].actions.length; actionIndex++) {
            addAction(transAction.steps[stepIndex].actions[actionIndex].type);
        }
    }
    enableActionDetail();
}

function onRecordPressed() {
    try{
        if (isRecording) {
            isRecording = false;
            removeEventListener("DOMContentLoaded", onRecorderLoad, true);

            mainBox.removeChild(contentIFrame);
            transAction = OneAPMRecorder.stopRecording();

            didInitRecorder = false;
            RecButton.label = "录制";
            RecButton.image = "chrome://oneapmrecorder/skin/icons/record.png"
            OpenButton.disabled = false;
            if (stepQueue.childNodes.length>0) {
                SaveButton.disabled = false;
                enableActionDetail();
                PlayButton.disabled = false;
            }
        } else {
            if (stepQueue.childNodes.length>0 && !didSave) {
                if (!window.confirm("脚本还未保存，重新开始录制将丢失之前录制的操作！")) {
                    return;
                }
            }
            isRecording = true;
            RecButton.label = "停止";
            RecButton.image = "chrome://oneapmrecorder/skin/icons/stopR.png"
            SaveButton.disabled = true;
            OpenButton.disabled = true;
            PlayButton.disabled = true;

            OneAPMRecorder.startRecording();
            initRecorder();
        }
    } catch ( ex) {
        logException( ex , "onRecordPressed");
    }
}

function breakRecord() {
    if (isRecording) {
        onRecordPressed();
        alert("操作数量已达到最大数量限制，录制自动停止！");
    }
}

function checkStepTitle() {
    if (transAction) {
        for (var stepIndex = 0; stepIndex < transAction.steps.length; stepIndex++) {
            if (transAction.steps[stepIndex].title.length == 0) {
                var text = "第"+ (stepIndex+1) +"步标题不能为空！"
                alert(text);
                return true;
            }
        }
    }
    return false;
}

function onSavePressed() {
    try {
        if (checkStepTitle()) return;
        var mode = Ci.nsIFilePicker.modeSave;
        var fp = Cc["@mozilla.org/filepicker;1"].createInstance(Ci.nsIFilePicker);
        fp.init(window, "Save Script", mode);
        fp.appendFilter("JSON", "*.json");
        fp.defaultExtension = '*.json';

        fp.defaultString =transAction.name;

        fp.open(function() {
            if(fp.file && fp.file.path) {
                var savePath = fp.file.path;
                var ext =  fp.file.path.match(/[.][^.]*$/);
                if (ext[0].toLowerCase() !== '.json') {
                    savePath += '.json';
                }

                var needToConfirm = savePath !== fp.file.path;
                if (openFile(savePath).exists() && needToConfirm) {
                    if (window.confirm("文件 " + savePath + " 已存在。 是否覆盖？")) {
                        saveScript(savePath);
                    }
                } else {
                     saveScript(savePath);
                }
            }
            this.cancel();
        });
    }catch(ex){
           logException(ex, "onSavePressed");
    }
}

function onOpenPressed() {
    try {
        if (stepQueue.childNodes.length>0 && !didSave) {
            if (!window.confirm("脚本还未保存，打开新的脚本将丢失之前录制的操作！")) {
                return;
            }
        }

        var mode = Ci.nsIFilePicker.modeOpen;
        var fp = Cc["@mozilla.org/filepicker;1"].createInstance(Ci.nsIFilePicker);
        fp.init(window, "Open Script", mode);
        fp.appendFilter("JSON", "*.json");
        fp.defaultExtension = '*.json';

        fp.open(function() {
            if (fp.file && fp.file.path) {
                var fileData = readFile(openFile(fp.file.path))
                fileData = fileData.replace(/: \[ - ([0-9]),/g, ': \[-$1,');
                transAction = JSON.parse(fileData);

                if (transAction) {
                    showTransAction();
                    if (mainBox.lastChild) recycleBin.appendChild(mainBox.lastChild);
                    didSave = true;
                    PlayButton.disabled = false;
                    SaveButton.disabled = false;
                }
            }
            this.cancel();
        })
    }catch(ex) {
           logException(ex, "onOpenPressed");
    }
}

function onPlayPressed() {
    try {
        if (isPlayBacking) {
            isPlayBacking = false;
            setActionEnable(true);
            RecButton.disabled = false;
            OpenButton.disabled = false;
            SaveButton.disabled = false;
            PlayButton.image = "chrome://oneapmrecorder/skin/icons/play.png";
            PlayButton.label = "播放";
            removeEventListener("DOMContentLoaded", onPlayerLoad, true);

            executor.stopPlay();
            mainBox.removeChild(playIFrame);
        } else {
            isPlayBacking = true;
            setActionEnable(false);
            RecButton.disabled = true;
            OpenButton.disabled = true;
            SaveButton.disabled = true;
            PlayButton.image = "chrome://oneapmrecorder/skin/icons/stopP.png"
            PlayButton.label = "停止";
            if (mainBox.lastChild) recycleBin.appendChild(mainBox.lastChild);

            initPlayer();
        }
    }catch(ex) {
           logException(ex, "onPlayPressed");
    }
}

function initPlayer() {
    getLogger().debug("initPlayer.");
    try{
        playIFrame = document.createElementNS(XULNS, "iframe");
        playIFrame.setAttribute("flex",1);
        playIFrame.style.position = "relative";
        addEventListener("DOMContentLoaded", onPlayerLoad, true);
        playIFrame.setAttribute("src", "chrome://oneapmrecorder/content/playBackingBrowser.xul");
        mainBox.appendChild(playIFrame);
    }catch(ex){
        logException( ex , "initPlayer");
    }
}

function onPlayerLoad(evt) {
    try{
        if (evt.originalTarget == playIFrame.contentDocument) {
            getLogger().debug("onPlayerLoad.");
            tabBox = playIFrame.contentDocument.getElementById("tab-box");
            tabBox.tabs.onselect = function() {
                tabBox.selectedTab.browser.updateUrl();
            }
            var cookieMgr = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager);
            cookieMgr.removeAll();

            executor = new scriptExecutor();
            resetRecorder();

            playIFrame.contentDocument.removeEventListener("DOMContentLoaded", onRecorderLoad, true);
            var result = playBack();
            if (isPlayBacking) {
                alert(result);
                onPlayPressed();
            }
        }
    }catch(ex) {
        logException( ex , "onPlayerLoad");
    }
}

function createFile(file) {
    try {
        file.create(0x00, 0664);
        return true;
    }
    catch(e) {
        return false;
    }
}

function openFile(path) {
    try {
        var file = Cc['@mozilla.org/file/local;1'].createInstance(Ci.nsILocalFile);
        file.initWithPath(path);
        return file;
    }
    catch(e) {
        return false;
    }
}

function readFile(file) {
    try {
        var data = new String();
        var fiStr = Cc['@mozilla.org/network/file-input-stream;1'].createInstance(Ci.nsIFileInputStream);
        var siStr = Cc['@mozilla.org/scriptableinputstream;1'].createInstance(Ci.nsIScriptableInputStream);
        fiStr.init(file, 1, 0, false);
        siStr.init(fiStr);
        data += siStr.read(-1);
        siStr.close();
        fiStr.close();
        data = this.toUnicode('UTF-8', data);
        return data;
    } catch(e) {
        return false;
    }
}

function writeFile(file, data, mode) {
    try {
        var foStream = Cc['@mozilla.org/network/file-output-stream;1'].createInstance(Ci.nsIFileOutputStream);
        data = this.fromUnicode( 'UTF-8', data);
        var flags = 0x02 | 0x08 | 0x20; // wronly | create | truncate
        foStream.init(file, flags, 0664, 0);
        foStream.write(data, data.length);
        foStream.close();
        return true;
    } catch(e) {
        return false;
    }
}

function fromUnicode(charset, data) {
    try {
        var uniConv = Cc['@mozilla.org/intl/scriptableunicodeconverter'].createInstance(Ci.nsIScriptableUnicodeConverter);
        uniConv.charset = charset;
        data = uniConv.ConvertFromUnicode(data);
    }
    catch(ex) {
        logException( ex , "fromUnicode");
    }
    return data;
}

function toUnicode(charset, data) {
    try{
        var uniConv = Cc['@mozilla.org/intl/scriptableunicodeconverter'].createInstance(Ci.nsIScriptableUnicodeConverter);
        uniConv.charset = charset;
        data = uniConv.ConvertToUnicode(data);
    } catch(ex) {
        logException( ex , "toUnicode");
    }
    return data;
}

function saveFile(path, data) {
    var file = openFile(path);
    if (file.exists()) {
        getLogger().debug("file exists, removing file.");
        file.remove(false);
    }

    if (!createFile(file)) {
        getLogger().error("failed to create " + path);
        return false;
    }

    writeFile(file, data);
    return true;
}

function saveScript(path) {
    try {
        var r;
        if (transAction) {
            var eof = "\r\n\r\n";
            r = saveFile(path, JSON.stringify(transAction) + eof);
        } else {
            r = false;
        }
    } catch(ex) {
        logException(ex, "saveScript");
    }

   if (r) {
        getLogger().debug("done writing: " + path);
        didSave = true;
    } else {
        alert("脚本保存失败。");
    }
}

function openScript(path) {
    try {
    } catch(ex) {
        logException(ex, "openScript");
    }
}

function CreateTab(aURI) {
    try {
        var newTab = document.createElementNS(XULNS, "tab");

        var isBlank = (aURI == undefined || aURI == "about:blank");

        if (isBlank)
            newTab.setAttribute("label", "");
        else
            newTab.setAttribute("label", aURI);

        newTab.maxWidth = 150;
        newTab.minWidth = 60;
        newTab.width = 0;
        newTab.setAttribute("flex", "1");
        newTab.setAttribute("crop", "end");
        newTab.className = "browser-tab";

        newTab.browser = newBrowser = document.createElementNS(XULNS, "browser");
        newBrowser.setAttribute("type", "content-targetable");
        newBrowser.setAttribute("flex", "1");

        newBrowser.addEventListener("DOMTitleChanged", function (evt) {
                if (evt.target == this.contentDocument) {
                    newTab.setAttribute("label", newTab.browser.contentTitle);
                    newTab.setAttribute("crop", "end");
                }
            }, true);

        newBrowser.updateUrl = function() {
            if (tabBox.selectedTab == newTab) {
                if (tabBox.selectedTab.browser.currentURI)
                    contentIFrame.contentDocument.getElementById("location-box").value = tabBox.selectedTab.browser.currentURI.spec;
            }
        }

        tabBox.tabs.appendChild(newTab);
        tabBox.tabpanels.appendChild(newBrowser);

        if (!isBlank) {
            newBrowser.stop();
            newBrowser.userTypedValue = aURI;

            OneAPMRecorder.attachBrowser(newBrowser);

            if (isPlayBacking) {
                executor.add(newBrowser);
            } else {
                newBrowser.loadURI(aURI);
            }
        } else {
            if (isPlayBacking) {
                executor.add(newBrowser);
            } else {
                OneAPMRecorder.attachBrowser(newBrowser);
            }
        }

        newBrowser.addEventListener('DOMContentLoaded', function(e) {
            var doc = e.target,
            win = doc.defaultView.wrappedJSObject,
            busyWait = true,
            timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer),
            handleTime = {
                notify: function() {
                    if(!busyWait) {
                        timer.cancel();
                    }
                }
            };

            win.showModalDialog = function(url, args, opts) {
                var modalWin = win.open(url, opts);
                modalWin.close = function(e) {
                    busyWait = false;
                }
                modalWin.dialogArguments = args;

                timer.initWithCallback(handleTime, 1000, Ci.nsITimer.TYPE_REPEATING_SLACK);

                while (busyWait){
                    Services.tm.currentThread.processNextEvent(true);
                }

                return modalWin.returnValue;
            }
        }, false);
    }catch(ex){
        logException( ex , "CreateTab");
    }

    return newTab;
}

function nsBrowserAccess() {
}

nsBrowserAccess.prototype = {
    QueryInterface : function(aIID)
    {
        if (aIID.equals(Ci.nsIBrowserDOMWindow) || aIID.equals(Ci.nsISupports)) {
            return this;
        }
        throw Cr.NS_NOINTERFACE;
    },

    openURI : function(aURI, aOpener, aWhere, aContext)
    {
        var isExternal = (aContext == Ci.nsIBrowserDOMWindow.OPEN_EXTERNAL);
        newWindow = CreateTab(aURI).browser.docShell.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);
        if (aURI) {
            if (aOpener) {
                location = aOpener.location;
                referrer = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService).newURI(location, null, null);
            }
            newWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIWebNavigation).loadURI(aURI.spec, Ci.nsIChannel.INHIBIT_CACHING | Ci.nsIChannel.LOAD_BYPASS_CACHE, null, null, null);
        }
        if (isExternal)
            newWindow.focus();

        return newWindow;
    },

    isTabContentWindow : function(aWindow)
    {
        // Shouldn't ever get called
        throw Cr.NS_ERROR_UNEXPECTED;
    }
};

function resetRecorder() {
    getLogger().debug("resetRecorder.");
    try{
        var tabpanels = tabBox.lastChild;
        while (tabpanels.childNodes.length > 0) tabpanels.removeChild(tabpanels.firstChild);

        // remove the panels first, since there may be popups as we try to leave the page
        var tabs = tabBox.firstChild;
        while (tabs.childNodes.length > 0) {
            OneAPMRecorder.detachBrowser(tabs.firstChild.browser);

            tabs.removeChild(tabs.firstChild);
        }

        browser = CreateTab().browser;
        browser.setAttribute("class", "content-primary");
        browser.setAttribute("src", "about:blank");
        tabBox.selectedIndex = 0;

        window.QueryInterface(Ci.nsIDOMChromeWindow).browserDOMWindow = new nsBrowserAccess(browser);

        browser.mPanelContainer = {
            childNodes : []
        };
        browser.mTabs = [];
        browser.browsers = [];
        browser.mTabBox = { selectedIndex : 0 };
        browser.attachFormFill = function () {};
    }catch(ex){
        logException( ex , "resetRecorder");
    }
}

function onRecorderLoad(evt) {
    try{
        if (evt.originalTarget == contentIFrame.contentDocument) {
            tabBox = contentIFrame.contentDocument.getElementById("tab-box");
            tabBox.tabs.onselect = function() {
                tabBox.selectedTab.browser.updateUrl();
            }

            resetRecorder();

            contentIFrame.contentDocument.removeEventListener("DOMContentLoaded", onRecorderLoad, true);
        }
    }catch(ex) {
        logException( ex , "onRecorderLoad");
    }
}

function initRecorder() {
    getLogger().debug("initRecorder.");
    try{
        if (didInitRecorder) return;

        while (stepQueue.childNodes.length > 0) stepQueue.removeChild(stepQueue.firstChild);

        contentIFrame = document.createElementNS(XULNS, "iframe");
        contentIFrame.setAttribute("flex",1);
        contentIFrame.style.position = "relative";
        addEventListener("DOMContentLoaded", onRecorderLoad, true);
        contentIFrame.setAttribute("src", "chrome://oneapmrecorder/content/recordingBrowser.xul");
        if (mainBox.lastChild) recycleBin.appendChild(mainBox.lastChild);
        mainBox.appendChild(contentIFrame);

        didInitRecorder = true;
        didSave = false;
    }catch(ex){
        logException( ex , "initRecorder");
    }
}

function getTargetNameByWindow(win) {
    var name = "";
    var tabs = contentIFrame.contentDocument.getElementById("tab-bar");
    for (var i = 0; i<tabs.childNodes.length; i++) {
        var tab = tabs.getItemAtIndex(i);
        if (tab.browser.contentWindow == win) {
            name = "top["+i+"]";
            break;
        }
    }
    return name; 
}

var OneAPMRecorder = {
    session : null,

    startRecording : function() {
        LocatorGenerator.init($, getTargetNameByWindow);

        this.session = new OneApmBrowserEventCenter(window, actionMaxNum);
        this.session.badCertListener = window.badCertListener.prototype;
        var cacheMgr  = Cc["@mozilla.org/network/cache-service;1"].getService(Ci.nsICacheService);
        var cookieMgr = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager);
        try {
            //cacheMgr.evictEntries(Ci.nsICache.STORE_ON_DISK);
            //cacheMgr.evictEntries(Ci.nsICache.STORE_IN_MEMORY);
            cookieMgr.removeAll();
        } catch (e) {
            getLogger().error("clear cookies failed");
        }
    },

    stopRecording : function() {
        return this.detachAllBrowsers();
    },

    attachBrowser : function(browser) {
        if (!this.session) return;
        if(browser) {
            getLogger().debug("attaching browser into session");
            this.session.attach(browser);
        }
    },

    detachBrowser : function(browser) {
        if(browser) {
            this.session.detach(browser);
        }
    },

    detachAllBrowsers : function() {
        var transa = this.session.finish();
        this.session.detachAll();
        return transa;
    }
};

var myListener =
{
    QueryInterface: function(iid)
    {
        if (iid.equals(Ci.nsIURIContentListener) || iid.equals(Ci.nsISupportsWeakReference) || iid.equals(Ci.nsISupports))
            return this;
        throw Cr.NS_NOINTERFACE;
    },
    onStartURIOpen: function(aUri)
    {
           return false;
    },
    doContent: function(aContentType, aIsContentPreferred, aRequest, aContentHandler )
    {
        getLogger().debug("docontent");
        throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    },
    canHandleContent: function(aContentType, aIsContentPreferred, aDesiredContentType)
    {
        getLogger().debug("canhandle");
        throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    },
    isPreferred: function(aContentType, aDesiredContentType)
    {
        try
        {
            getLogger().debug("ispreferred");
            var webNavInfo = Cc["@mozilla.org/webnavigation-info;1"].getService(Ci.nsIWebNavigationInfo);
            return webNavInfo.isTypeSupported(aContentType, null);
        }
        catch (e)
        {
            return false;
        }
    },
    GetWeakReference : function()
    {
        getLogger().debug("GetWeakReference");

        throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    }
}

var wnd = window.QueryInterface(Ci.nsIInterfaceRequestor)
                        .getInterface(Ci.nsIWebNavigation)
                        .QueryInterface(Ci.nsIDocShell)
                        .QueryInterface(Ci.nsIInterfaceRequestor)
                        .getInterface(Ci.nsIURIContentListener);
wnd.parentContentListener = myListener;

function badCertListener() {}
badCertListener.prototype = {
    create: function() {
        return new badCertListener();
    },
    getInterface: function (aIID) {
        return this.QueryInterface(aIID);
    },
    QueryInterface: function(aIID) {
        if (aIID.equals(Components.interfaces.nsIBadCertListener2) ||
            aIID.equals(Components.interfaces.nsIInterfaceRequestor) ||
            aIID.equals(Components.interfaces.nsISupports))
            return this;
        throw Components.results.NS_ERROR_NO_INTERFACE;
    },
    notifyCertProblem: function MSR_notifyCertProblem(socketInfo, sslStatus, targetHost) {
        getLogger().debug ("notifyCertProblem, adding exception if possible: targetHost: "+targetHost);
        this.addException(sslStatus.QueryInterface(Components.interfaces.nsISSLStatus).serverCert,
                          sslStatus, targetHost);
        return true; // suppress error UI
    },
    /**
     * Handle user request to add an exception for the specified cert
     */
    addException: function(cert, sslStatus, hoststring) {
        if(!cert || !sslStatus) {
            getLogger().debug("cert or sslStatus was null: cert: "+cert+" sslStatus: "+sslStatus);
            return;
        }

        var overrideService = Components.classes["@mozilla.org/security/certoverride;1"]
                                .getService(Components.interfaces.nsICertOverrideService);
        var flags = 0;
        if(sslStatus.isUntrusted)
            flags |= overrideService.ERROR_UNTRUSTED;
        if(sslStatus.isDomainMismatch)
            flags |= overrideService.ERROR_MISMATCH;
        if(sslStatus.isNotValidAtThisTime)
            flags |= overrideService.ERROR_TIME;

        var shouldStorePermanently = false;

        var uri = this.getURI(hoststring);
        getLogger().debug("adding cert override for: "+uri.asciiHost + ":"+uri.port+", cert: "+cert.toString()+", flags: "+flags);
        overrideService.rememberValidityOverride(
            uri.asciiHost, uri.port,
            cert,
            flags,
            !shouldStorePermanently);
        getLogger().debug("reloading entire browser frame");
        contentIFrame.contentWindow.navigateButton(true);
    },
    /**
     * Build and return a URI, based on the information supplied in the
     * Certificate Location fields
     */
    getURI: function getURI(hoststring) {
      var fus = Components.classes["@mozilla.org/docshell/urifixup;1"].getService(Components.interfaces.nsIURIFixup);
      var uri = fus.createFixupURI(hoststring, 0);

      if(!uri) return null;

      if(uri.scheme == "http") uri.scheme = "https";

      if (uri.port == -1) uri.port = 443;

      return uri;
    }
}