Cu.import("resource://gre/modules/XPCOMUtils.jsm");
const WaitStatus = {"StatusOK":0,"StatusTimeout":1,"StatusError":2};
const observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
const windowMediator = Cc['@mozilla.org/appshell/window-mediator;1'].getService(Ci.nsIWindowMediator);
const nsIHttpActivityObserver = Ci.nsIHttpActivityObserver;
const activityDistributor = Cc["@mozilla.org/network/http-activity-distributor;1"].getService(Ci.nsIHttpActivityDistributor);
const WAIT_SLICE = 250;
const COMPENSATION_FOR_CHECKING_HARDDEADLINE = 500;

function playBack() {
    var started = false;
    var running = true;
    var currentStepIdx = 0;
    var currentStep = transAction.steps[currentStepIdx];
    ONEAPM_CURRENT_STEP_INDEX = 0;
    ONEAPM_CURRENT_ACTION_INDEX = 0;

    while(running) {
        ONEAPM_CURRENT_STEP_INDEX = currentStepIdx;
        getLogger().debug("Running step " + currentStepIdx + " with " + currentStep.actions.length + " actions");

        if (!started) {
            started = true;
        }

        for (var i=0; i<currentStep.actions.length; ++i) {
            try {
                ONEAPM_CURRENT_ACTION_INDEX = i;
                var actionJson = currentStep.actions[i];
                if (actionJson.isDone) {
                    getLogger().debug("Action is already done, step: " + currentStepIdx + " act: " + i);
                    continue;
                }
                var actionObj = interpretAction(actionJson);
                actionObj.execute();

            } catch(ex) {
                getLogger().error("Caught exception running test: " + ex.message);
                var errorMsg = "脚本播放失败：于第 " + (currentStepIdx+1) + " 步，第 " + (i+1) + "个操作，";
                if (ex.message == 'TIMEOUT_EXCEEDED') {
                    errorMsg += "等待超时！";
                } else if (ex.message == 'INTERNAL_ERROR') {
                    errorMsg += "执行失败！";
                }
                return errorMsg;
            }
        }

        currentStepIdx++;
        if (currentStepIdx >= transAction.steps.length) running = false;
        else currentStep = transAction.steps[currentStepIdx];
    }
    return "脚本播放成功";
}

function enhanceWindow(win) {
    win = win.wrappedJSObject || win;
    win.oneapm_execScript = function(script) {
        injectScriptWithTag(script, win);
    };
    return win;
}

function injectScriptWithTag(script, window) {
    if(!window || !window.document || !window.document.body) {
        throw "The window must be a valid DOM window with a valid document and body";
    }
    var injected = window.document.createElement('script');
    injected.setAttribute('type', 'text/javascript');
    injected.setAttribute('id', 'ONEAPM_SIZZLE_SCRIPT_ELEMENT');
    script = "var _gozilla_inject_user_code_ = function (){ " + script + "};" ;
    script = script + "try { _gozilla_inject_user_code_(); } catch( ex ) { var message = 'user_script_error#' + ex;alert( message); } ";
    script = "( function() { " + script + " } ) () ;" ;
    var tn = window.document.createTextNode(script);
    injected.appendChild(tn);
    window.document.body.appendChild(injected);
    injected = window.document.getElementById( "ONEAPM_SIZZLE_SCRIPT_ELEMENT" )
    if ( injected ) {
        injected.parentNode.removeChild( injected );
    }
}

function hadInternalError() {
    return false;
}

function interpretAction(actionJson) {
    //getLogger().debug("interpretAction start, actionJson= " + JSON.stringify(actionJson));
    if (actionMap[actionJson.type] == undefined) {
        getLogger().debug("Unknown action: " + actionJson.type);
        return new Action(actionJson);
    } else {
        return new actionMap[actionJson.type](actionJson);
    }
}

function Action() {}
Action.prototype = {
    execute: function() {
        getLogger().debug("Execute action");
    }
}

actionMap = {};

function Command() {}
Command.prototype = new Action();
Command.prototype.toJS = function() {};

function NavigateCommand(params) {
    this.params = params;
}
NavigateCommand.prototype = new Command();
NavigateCommand.prototype.execute = function() {
    getLogger().debug("Navigating window");
    
    try{
        var win = executor.browsers[0].browser.contentWindow;
        executor.navigateFrame(win, this.params.url);
    }catch(ex){
        getLogger().error("Captured exception while navigating, "+ ex.message);
    }
}
actionMap["navigate"] = NavigateCommand;

function Wait(params) {
    this.params = params;
}
Wait.prototype = new Action();
Wait.prototype.execute = function() {

    try {
        if (this.params.style == "time" ) {
            var time = parseInt(this.params.timeout);
            getLogger().debug("Waiting for " + time + "ms");
            var completed = executor.waitFor(time);
            if (!completed) {
                getLogger().debug("Got internal error during wait for time");
                throw (new Error("INTERNAL_ERROR"));
            }
        } else if (this.params.style == "page_complete" ) {
            getLogger().debug("Waiting for page load to complete");
            executor.waitForNetRequestsComplete(1000, WAIT_SLICE);
            var maximumWait = this.params.timeout ? parseInt(this.params.timeout) : 60*1000;
            var endTime = new Date().getTime() + maximumWait;
            var pageCompleteDelay = this.params.pageCompleteDelay ? parseInt(this.params.pageCompleteDelay) : -1;
            var pendingReadWait = this.params.pendingReadWait ? parseInt(this.params.pendingReadWait) : -1;
            getLogger().debug("Waiting for page load to complete, maximumWait: " + maximumWait + ", pageCompleteDelay: " + pageCompleteDelay + ", pendingReadWait: " + pendingReadWait);
            var rvalue = executor.waitForPageComplete(maximumWait , pageCompleteDelay , pendingReadWait );

            if (rvalue == 1) {
                throw (new Error("TIMEOUT_EXCEEDED"));
            } else if (rvalue == 2) {
                getLogger().debug("Got internal error during wait for page complete");
                throw (new Error("INTERNAL_ERROR"));
            }
            executor.waitForNetRequestsComplete(2000, WAIT_SLICE);
        } else if (this.params.style == "network" ) {
            var maximumWait = this.params.timeout ? parseInt(this.params.timeout) : 60*1000;
            var endTime = new Date().getTime() + maximumWait;
            getLogger().debug("Waiting for network operations to complete for a maximum of " + maximumWait + "ms");
            var rvalue = executor.waitForNetRequestsComplete(maximumWait, 200);
            getLogger().debug("Got return value of " + rvalue + " from native wait for net call");

            if (rvalue == 1) {
                var currentTime = new Date().getTime();
                if ( (currentTime + COMPENSATION_FOR_CHECKING_HARDDEADLINE) >= endTime ){
                    throw (new Error("TIMEOUT_EXCEEDED"));
                }
            } else if (rvalue == 2) {
                getLogger().debug("Got internal error during wait for net");
                throw (new Error("INTERNAL_ERROR"));
            }
        }
    }  finally {
        getLogger().debug("----endWaitAction----");
    }

}
actionMap["wait"] = Wait;

function ClickCommand(params) {
    this.locator = new Locator(params);
    this.params = params;
}
ClickCommand.prototype = new Command();
ClickCommand.prototype.execute = function () {
    var el = this.locator.execute();
    var elPos = findPos(el);
    var cx = elPos[0] + el.clientWidth / 2;
    var cy = elPos[1] + el.clientHeight / 2;
    var coords = { "clientX" : cx, "clientY" : cy };

    fireEvent('mousemove', el, coords);
    if (el.getAttribute("type") != "application/x-shockwave-flash") {
        fireEvent('mouseover', el, coords);
        fireEvent('mousedown', el, coords);
        fireEvent('focus', el, coords);
        fireEvent('mouseup', el, coords);
    }
    if (this.params.doubleclick) {
        fireEvent('dblclick', el, coords);
    } else {
        fireEvent('click', el, coords);
    }
    try {
        fireEvent('mouseout', el, coords);
    } catch (ex) {
        getLogger().error("get clcik exeception : " + ex.message);
    }
}
actionMap["click"] = ClickCommand;

function keytInCommand(params) {
    this.locator = new Locator(params);
    this.params = params;
    if (this.params.textValue == undefined) {
        this.params.textValue = '';
    }
}
keytInCommand.prototype = new Command();

keytInCommand.prototype.typeString = function(currentString, idx, endIdx, newPart) {
    if (idx == endIdx && idx >= currentString.length-1) return currentString + newPart;
    return currentString.slice(0, idx) + newPart + currentString.slice(endIdx, currentString.length);
}
keytInCommand.prototype.deleteCharAt = function(currentString, idx) {
    if (idx <= 0) return "";
    return currentString.slice(0, idx-1) + currentString.slice(idx, currentString.length);
}
keytInCommand.prototype.execute = function() {
    var el = this.locator.execute();

    var isTextEl = (el.tagName == "INPUT" && (el.type == 'email' || el.type == 'tel' || el.type == 'search' ||
        el.type == 'url' || el.type == 'number' || el.type == 'range' || el.type == 'time' || el.type == 'month' ||
        el.type == 'week' || el.type == 'date' || el.type == 'datetime' || el.type == 'datetime-local') ||
        (el.type == 'text') || (el.type == 'password')) || el.tagName == "TEXTAREA";

    var isMeta = false, isCtrl = false, isShift = false, isAlt = false;
    for (var m=0; this.params.modifiers && m<this.params.modifiers.length; m++) {
        var mod = this.params.modifiers[m];
        if (mod == "META") isMeta = true;
        else if (mod == "SHIFT") isShift = true;
        else if (mod == "ALT") isAlt = true;
        else if (mod == "CTRL") isCtrl = true;
    }

    for (var i=0; i<this.params.charCodes.length; ++i) {
        var preventDefault = false;
        var evtArgs = { charCode : this.params.charCodes[i], keyCode : this.params.keyCodes[i],
            ctrlKey:isCtrl, altKey:isAlt, shiftKey:isShift, metaKey:isMeta };
        if (this.params.keyCodes[i] == 13 && isTextEl && el.tagName != 'TEXTAREA') {
            if ( i == this.params.charCodes.length-1 && el.value != this.params.textValue ) {
                el.value = this.params.textValue;
            }
            preventDefault = true;
        }
        fireEvent('keydown', el, evtArgs, preventDefault);
        if (isTextEl && !isMeta && !isAlt && !isCtrl && (this.params.keyCodes[i] == 32 || this.params.keyCodes[i] > 46)) {
            el.value = this.typeString(el.value, this.params.selectionStart[i], this.params.selectionEnd[i],
                String.fromCharCode((this.params.charCodes[i] != 0 ? this.params.charCodes[i] : this.params.keyCodes[i])));
        } else if (isTextEl && this.params.keyCodes[i] == 8) {
            el.value = this.deleteCharAt(el.value, this.params.selectionStart[i]);
        } else if (isTextEl && this.params.keyCodes[i] == 46) {
            el.value = this.typeString(el.value, this.params.selectionStart[i], this.params.selectionEnd[i], '');
        } else if (this.params.keyCodes[i] == 13) {
            if (el.tagName == 'TEXTAREA') {
                el.value = el.value + "\r\n";
            }
        }

        if (isTextEl && i == this.params.charCodes.length-1 && el.value != this.params.textValue) {
            el.value = this.params.textValue;
        }

        fireEvent('keyup', el, evtArgs, preventDefault);
    }

    if (isTextEl) {
        fireEvent('change', el, {});
        fireEvent('input', el, {});
    }
}
actionMap["keytIn"] = keytInCommand;

function SelectOptionCommand(params) {
    this.locator = new Locator(params);
    this.params = params;
}
SelectOptionCommand.prototype = new Command();
SelectOptionCommand.prototype.findOptionIndex = function(sel, optionTxt) {
    for (var i=0; i<sel.options.length; ++i) {
        if (sel.options[i].textContent.indexOf(optionTxt) != -1) {
            getLogger().debug("Found index " + i + " for option with text " + optionTxt);
            return i;
        }
    }
    return -1;
}
SelectOptionCommand.prototype.execute = function() {
    if (this.params.optionIndex != undefined && this.params.optionIndexes == undefined) {
        this.params.optionIndexes = [];
        this.params.optionIndexes[0] = this.params.optionIndex;
    }
    if (this.params.textValue != undefined && this.params.textValues == undefined) {
        this.params.textValues = [];
        this.params.textValues[0] = this.params.textValue;
    }

    var sel = this.locator.execute();
    sel.selectedIndex = -1;

    for (var i=0; i<this.params.optionIndexes.length; ++i) {
        var idx = -1;
        if (this.params.textValues.length >= i) {
            idx = this.findOptionIndex(sel, this.params.textValues[i]);
        }
        if (idx == -1) idx = this.params.optionIndexes[i];
        getLogger().debug("Selecting form option " + idx + " from select " + this.locator);
        if (sel.selectedIndex == -1 || idx < sel.selectedIndex)
            sel.selectedIndex = idx;
        var opt = sel.options[idx];
        opt.selected = true;
        fireEvent('click', opt, {"clientX": 0, "clientY": 0});

        if (!this.params.suppressEvents) {
            fireEvent('change', sel, {});
        }
    }
}
actionMap["selectOption"] = SelectOptionCommand;

function findPos(obj) {
    var rect = obj.getBoundingClientRect();
    return [rect.left, rect.top];
}

function fireEvent(type, srcEl, options, preventDefault) {
    //getLogger().debug("fireEvent start {");
    var mouse =  (/^mouse(over|out|down|up|move)|(dbl)?click$/.test(type));
    var keyboard = !mouse && (/^key(up|down|press)$/.test(type));

    var evt;
    if (mouse) {
        var e = extend({
            bubbles: true, cancelable: (type != "mousemove"), view: window, detail: 0,
            screenX: 0, screenY: 0, clientX: 0, clientY: 0,
            ctrlKey: false, altKey: false, shiftKey: false, metaKey: false,
            button: 0, relatedTarget: undefined
        }, options);

        if ( type.indexOf("click") >= 0 && (checkLocalName( srcEl, "A" ) ||
                                            checkLocalName( srcEl.parentNode , "A" ) ||
                                            checkLocalName( srcEl ,  "IMG" ) || 
                                            checkLocalName( srcEl ,  "AREA") ) ) {
            //getLogger().debug("srcEl = new XPCNativeWrapper(srcEl);");
            srcEl = new XPCNativeWrapper(srcEl);
        }
        //getLogger().debug("evt = srcEl.ownerDocument.createEvent(\"MouseEvents\");");
        evt = srcEl.ownerDocument.createEvent("MouseEvents");
        //getLogger().debug("evt.initMouseEvent('"+type+"',true,"+e.cancelable+
        //    ",srcEl.ownerDocument.defaultView,"+e.detail+","+e.screenX+","+e.screenY+
        //    ","+e.clientX+","+e.clientY+","+e.ctrlKey+","+e.altKey+","+e.shiftKey+","+
        //    e.metaKey+","+e.button+","+e.relatedTarget+");");
        evt.initMouseEvent(type, true, e.cancelable, srcEl.ownerDocument.defaultView, 
            e.detail, e.screenX, e.screenY, e.clientX, e.clientY, 
            e.ctrlKey, e.altKey, e.shiftKey, e.metaKey, 
            e.button, e.relatedTarget);
    } else if (keyboard) {
        var e = extend({ bubbles: true, cancelable: true, view: window,
            ctrlKey: false, altKey: false, shiftKey: false, metaKey: false,
            keyCode: 0, charCode: 0
        }, options);

        //getLogger().debug("evt = srcEl.ownerDocument.createEvent(\"KeyEvents\")");
        evt = srcEl.ownerDocument.createEvent("KeyEvents");
        if(type=="keypress") {
            //getLogger().debug("evt.initKeyEvent ('"+type+"',"+e.bubbles+","+e.cancelable+
            //    ",srcEl.ownerDocument.defaultView,"+e.ctrlKey+","+e.altKey+","+e.shiftKey+","+e.metaKey+","+((e.charCode > 0) ? 0: e.keyCode, e.charCode)+");");
            evt.initKeyEvent (type, e.bubbles, e.cancelable, srcEl.ownerDocument.defaultView, 
                e.ctrlKey, e.altKey, e.shiftKey, e.metaKey, 
                (e.charCode > 0) ? 0: e.keyCode, e.charCode);
            //getLogger().debug("srcEl.value += String.fromCharCode("+(e.charCode > 0 ? e.charCode : e.keyCode)+");");
            srcEl.value += String.fromCharCode( e.charCode > 0 ? e.charCode : e.keyCode );            
        }else if (type=="keyup" || (type=="keydown")){
            //getLogger().debug("evt.initKeyEvent('"+type+"',"+e.bubbles+","+e.cancelable+
            //    ",srcEl.ownerDocument.defaultView,"+e.ctrlKey+","+e.altKey+","+e.shiftKey+
            //    ","+e.metaKey+","+e.keyCode+",0);");
            evt.initKeyEvent (type, e.bubbles, e.cancelable, srcEl.ownerDocument.defaultView, 
                e.ctrlKey, e.altKey, e.shiftKey, e.metaKey, 
                e.keyCode, 0);
        }

        if (preventDefault) {
            //getLogger().debug("evt.preventDefault();");
            evt.preventDefault();
        }
        
    } else {
        //getLogger().debug("srcEl.ownerDocument.createEvent(\"UIEvents\");");
        evt = srcEl.ownerDocument.createEvent("UIEvents");
        //getLogger().debug("evt.initUIEvent('"+type+"', true, true, srcEl.ownerDocument.defaultView, 0);");
        evt.initUIEvent(type, true, true, srcEl.ownerDocument.defaultView, 0);
    }
    
    try {
        if (srcEl.replayEvent && (/(dbl)?click$|^key(down|up|press)|mousemove$/.test(type))) {
            //getLogger().debug("srcEl.replayEvent(evt);");
            srcEl.replayEvent(evt);
        }
        //getLogger().debug("srcEl.dispatchEvent(evt);");
        srcEl.dispatchEvent(evt);
    } catch(ex) {
        getLogger().error("Trapped failure running event handler: " + ex.message);
    }
    //getLogger().debug("} fireEvent end");
}

function superTypeOf(value) {
    var s = typeof value;
    if (s === 'object') {
        if (value) {
            if ( typeof value.length === 'number' && value.item == "[object]"){
                s = 'array';
            }
            else if ( typeof value.length === 'number' &&
                      typeof value.item === 'function' &&
                      value.toString() === '[object NodeList]' ){
                s = 'array';
            }
            else if (typeof value.length === 'number' &&
                !(value.propertyIsEnumerable('length')) &&
                typeof value.splice === 'function') {
                s = 'array';
            }
        } else {
            s = 'null';
        }
    }
    return s;
}

function extend(target, obj) {
    var target = arguments[0] || {}, i = 1, length = arguments.length, options;
    for ( ; i < length; i++ ) {
        if ( (options = arguments[ i ]) != null ) {
            for ( var name in options ) {
                var src = target[ name ], copy = options[ name ];
                if ( target === copy )
                    continue;
                else if ( copy !== undefined )
                    target[ name ] = copy;
            }
        }
    }
    return target;
}

function checkLocalName( aElement , aValue ) {
    return aElement && 
           aElement.localName && 
           aElement.localName.toUpperCase && 
           aElement.localName.toUpperCase() == aValue ;
}

function Locator(targetParams) {
    this.params = targetParams;
    this.locator4Page = null;
}
Locator.prototype.findFrame = function(targetWin, currentWin) {
    if (targetWin == "") return currentWin.wrappedJSObject || currentWin;
    
    if (currentWin == undefined) currentWin = executor.getMainWindow();
    var chunks = targetWin.split(".");
    var currentChunk = chunks.length == 0 ? targetWin : chunks[0];
    
    var remainingChunk = "";
    for (var i=1; i<chunks.length; ++i) {
        remainingChunk += chunks[i];
        if (i < chunks.length-1) remainingChunk += ".";
    }
    if (currentChunk == "top") {
        return this.findFrame(remainingChunk, executor.getMainWindow());
    } else if (currentChunk.indexOf("top") == 0) {
        var re = new RegExp("top\\[([0-9]*)\\]");
        var winIdx = re.exec(currentChunk)[1];
        if (winIdx == 0) return this.findFrame(remainingChunk, executor.getMainWindow());
        else {
            var popups = executor.getPopups();
            if (winIdx > popups.count()) throw (new Error("Window " + winIdx + " not found"));
            return this.findFrame(remainingChunk, popups.get(winIdx - 1));
        }
    } else if (currentChunk.indexOf("frames[") == 0) {
        var safeFramesArray = executor.getFrames(currentWin);
        getLogger().debug("The current win has " + safeFramesArray.count() + " frames");
        var re = new RegExp("frames\\[([0-9]*)\\]");
        var idx = re.exec(currentChunk)[1];
        currentWin = safeFramesArray.get(idx);
        return this.findFrame(remainingChunk, currentWin);
    } else {
        getLogger().error("Unknown chunk: " + currentChunk);
    }
    return currentWin.wrappedJSObject || currentWin;
};

Locator.prototype.loadSizzle = function(){
    var sizzeUrl = "resource://oneapmrecorder/sizzle.js"
    var sizzleBody = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest);
    sizzleBody.open("GET", sizzeUrl, false);
    sizzleBody.send(null);
    return sizzleBody.responseText;
};

Locator.prototype.injectSizzle = function(win) {
    getLogger().debug("loading sizzle");
    var body = this.loadSizzle();
    getLogger().debug("injecting sizzle");
    var oldGet = win.document.getElementsByClassName;
    if (oldGet) win.document.getElementsByClassName = undefined;
    win.oneapm_execScript(body);
    if (oldGet) win.document.getElementsByClassName = oldGet;
}
Locator.prototype.execute = function() {
    var win = executor.getMainWindow();
    if (this.params.window != undefined) {
        win = this.findFrame(this.params.window);
    }
    var el = null;
    for (var i=0; i<this.params.locators.length; ++i) {
        if (this.params.locators[i][0] == "css") {
            var cssSelector = this.params.locators[i][1];
            
            if (win.OneAPM_Sizzle == undefined) {
                try {
                    this.injectSizzle(win);
                    //make sure sizzle is ready
                    if(!win.OneAPM_Sizzle){
                        oneapmShimLog("OneAPM_Sizzle isn't ready right now, wait for it.");
                        var maxWaitForSizzle = 2000; //ms
                        var endTime = new Date().getTime() + maxWaitForSizzle;
                        while(!win.OneAPM_Sizzle){
                            if (new Date().getTime() > endTime) {
                                getLogger().error("failed to load sizzle within ms " + 2000);
                                //give chance to other potential dom selector
                                break;
                            }
                            executor.waitFor(25);
                        }
                    }
                } catch(ex) {
                    getLogger().error("Failed to inject sizzle : " + ex);
                    continue;
                }
            }
            try{
                var els = win.OneAPM_Sizzle(cssSelector);
                if (els.length == 0) continue;
                el = els[0];
                this.locator4Page = "OneAPM_Sizzle('" + cssSelector + "')[0]";
                break;
            }catch(ex) {
                getLogger().error("Got exception from within Sizzle while evaluating " + cssSelector);
            }
        } else if (this.params.locators[i][0] == "dom") {
            try {
                var domSelector = this.params.locators[i][1];
                var els = eval("win." + domSelector);
                this.locator4Page = domSelector;
            } catch(ex) {logException(ex,"execute"); continue;}
            if (els == null || superTypeOf(els) == "array") continue;
            el = els ;
            break;
        }
    }
    if (el == null) throw (new Error("Failed to locate object: " + this.toString()));
    getLogger().debug("Got " + el);
    el = el.wrappedJSObject || el;
    //it's experimental to highlight target element, it's useful for troubleshooting wrong css locator issues.
    //be careful because some element without style, e.g. document.
    if(el && el.style) el.style.border = "2px dotted yellow";
    return el;
};

Locator.prototype.getElements = function() {
    var win = executor.getMainWindow();
    if (this.params.window != undefined) {
        win = this.findFrame(this.params.window);
    }
    var el = null;
    for (var i=0; i<this.params.locators.length; ++i) {
        if (this.params.locators[i][0] == "css") {
            var cssSelector = this.params.locators[i][1];
            
            if (win.OneAPM_Sizzle == undefined) {
                try {
                    this.injectSizzle(win);
                    //make sure sizzle is ready
                    if(!win.OneAPM_Sizzle){
                        getLogger().debug("OneAPM_Sizzle isn't ready right now, wait for it.");
                        //wait for window.OneAPM_Sizzle.
                        var maxWaitForSizzle = 2000; //ms
                        var endTime = new Date().getTime() + maxWaitForSizzle;
                        while(!win.OneAPM_Sizzle){
                            if (new Date().getTime() > endTime) {
                                getLogger().error("failed to load sizzle within ms " + 2000);
                                //give chance to other potential dom selector
                                break;
                            }
                            executor.waitFor(25);
                        }
                    }
                } catch(ex) {
                    getLogger().error("Failed to inject sizzle : " + ex);
                    continue;
                }
            }
            try{
                var els = win.OneAPM_Sizzle(cssSelector);
                if (els.length == 0) continue;
                el = els;
                this.locator4Page = "OneAPM_Sizzle('" + cssSelector + "')[0]";
                break;
            }catch(ex) {
                logException(ex,"getElements");
            }
        } else if (this.params.locators[i][0] == "dom") {
            try {
                var domSelector = this.params.locators[i][1];
                var els = eval("win." + domSelector);
                this.locator4Page = domSelector;
            } catch(ex) {logException(ex,"getElements"); continue;}
            if (els == null ) continue;
            // we need to generate the real element array
            el = new Array();
            if (superTypeOf(els) == "array"){
                for( var nodeListIndex = 0 ; nodeListIndex < els.length ; ++ nodeListIndex ){
                    el.push( els[nodeListIndex] );
                }
            }else{
                el.push( els );
            }
            break;
        }
    }
    if (el == null) throw (new Error("Failed to locate object: " + this.toString()));
    getLogger().debug("Got " + el);
    return el;    
};

Locator.prototype.getLocator4Page = function(){
    return this.locator4Page;
}

Locator.prototype.toString = function() {
    if (this.params.locators == undefined || this.params.locators.length == 0) return "(no locator)"; 
    var windowName = this.params.window != undefined ? this.params.window : "top";
    if (this.params.locators[0][0] == "css")
        return "CSS selector " + this.params.locators[0][1] + " in window " + windowName;
    if (this.params.locators[0][0] == "dom")
        return "DOM selector " + this.params.locators[0][1] + " in window " + windowName;
    return "(unknown locator syntax)";
}

function scriptExecutor() {
    Cu.import("resource://oneapmrecorder/ComponentLogger.jsm");
    getLogger = ComponentLoggerFactory.getLoggerFunc("scriptExecutor");
    getLogger().debug("Constructoring scriptExecutor...");
    this.timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
}

scriptExecutor.prototype = {
    QueryInterface: XPCOMUtils.generateQI(),

    lastDataRead: new Date().getTime(),

    browsers: new Array(),

    hadInternalError: function() {
        return false;
    },

    stopPlay:function() {
        while(this.browsers.length > 0) {
            this.browsers[this.browsers.length - 1].stop();
            this.browsers[this.browsers.length - 1].uninstrument();
            this.browsers.pop();
        }
    },

    add: function(browser) {
        getLogger().trace("Adding browser.");
        if(!this.browsers.indexOf(browser) != -1) {
            this.browsers.push(new BrowserMonitor(browser, this));
            getLogger().trace("Added browser.");
        } else {
            getLogger().trace("Browser is already monitored.");
        }
    },

    navigateFrame: function(win, url) {
        getLogger().debug("navigate frame to url " + url);
        try{
            var uriFixUp = Cc["@mozilla.org/docshell/urifixup;1"].getService(Ci.nsIURIFixup);
            var urlFix = uriFixUp.createFixupURI(url, 0).spec;
            win.location = urlFix;
        }catch(ex) {
            getLogger().error("failed to navigating to " + url + " because of " + ex);
        }
    },

    observe: function(subject, topic, data) {
        try{
            if (topic == "timer-callback") {
                this.busyWait = false;
                this.timer.cancel();
            }else if(topic == "navigator_error"){
                getLogger().error("Navigator error happens.");
            }
        } catch(ex) {
            logException(ex,"observe");
        }
    },

    getMainWindow: function() {
        if((this.browsers.length != 0) && this.browsers[0] && this.browsers[0].browser && this.browsers[0].browser.contentWindow) {
            var mainWin = this.browsers[0].browser.contentWindow.QueryInterface(Ci.nsIDOMWindow);
            return enhanceWindow(mainWin);
        }else{
            getLogger().warn("main window isn't ready right now");
        }
    },

    getPopups: function() {
        var popups = [];
        if(this.browsers.length > 1) {
            for(var i=1;i<this.browsers.length;i++) {
                var contentWindow = null;
                if(this.browsers[i] && this.browsers[i].browser && this.browsers[i].browser.contentWindow) {
                    contentWindow = this.browsers[i].browser.contentWindow.QueryInterface(Ci.nsIDOMWindow);
                }
                popups.push(contentWindow);
            }
        }
        return {
            count: function() {return popups.length},
            get: function(index) {return  popups[index]==null?null:enhanceWindow(popups[index])}
        };
    },

    getFrames: function(win) {
        // cater to shim
        return {
            count: function() {return win.frames.length},
            get: function(index) {return enhanceWindow(win.frames[index])}
        };
    },

    doBusyWait: function(tm){
        if(!tm) return;
        this.busyWait = true;
        this.timer.init(this, tm, Ci.nsITimer.TYPE_ONE_SHOT);
        var thread = Cc["@mozilla.org/thread-manager;1"].getService(Ci.nsIThreadManager).currentThread;
        while (this.busyWait){
            thread.processNextEvent(true);
        }
    },

    waitFor: function(timeout) {
        getLogger().info("WaitFor " + timeout + " ms");
        var status = this.waitForTime(timeout);
        getLogger().info("WaitFor end without error ? " + status);
        return status;
    },

    waitForTime: function(timeout) {
        var hadError = false;
        while (timeout > 0 && !hadError){
            this.doBusyWait(WAIT_SLICE);
            hadError = this.hadInternalError();
            timeout -= WAIT_SLICE;
        }
        return hadError ? false : true;
    },

    waitForNetRequestsComplete: function(timeout, netRequestsCompleteThreshold) {
        try{
        getLogger().info("WaitForNetRequestsComplete,timeout: " + timeout + ",netRequestsCompleteThreshold: " + netRequestsCompleteThreshold);

        var status = WaitStatus.StatusOK;
        var timeoutRemaining = timeout > 1000? 1000: timeout;

        for (var i = 0; i < this.browsers.length; i++) {
            var startTime = new Date().getTime();
            var waitStatus = this.browsers[i].waitForNetRequestsComplete(
                    timeoutRemaining, netRequestsCompleteThreshold);

            if (waitStatus != WaitStatus.StatusOK) {
                getLogger().debug("waitForAllNetRequestsComplete fail with status " + waitStatus);
                return waitStatus;
            }

            var endTime = new Date().getTime();
            timeoutRemaining -= (endTime - startTime);
        }
        }catch(ex) {
            logException(ex,"waitForNetRequestsComplete");
        }

        return WaitStatus.StatusOK;
    },

    waitForPageComplete: function(aTimeout) {
        getLogger().debug("Wait For Page Complete " + aTimeout + " ms");
        var waitTime = aTimeout;
        var endTime = new Date().getTime() + waitTime;

        var status = WaitStatus.StatusOK;
        var timeoutRemaining = waitTime;
        for (var i = 0; i < this.browsers.length; i++) {
            var sTime = new Date().getTime();
            var status = this.browsers[i].waitForPageComplete(timeoutRemaining);

            if (status != WaitStatus.StatusOK) {
                getLogger().debug("waitForPageComplete for each browser fail with status "+ status);
                return status;
            }

            var eTime = new Date().getTime();
            timeoutRemaining -= (eTime - sTime);
        }

        waitTime = endTime - new Date().getTime();
        status = this.waitAfterPageComplete(waitTime);
        if (status != WaitStatus.StatusOK) {
            getLogger().debug("waitAfterPageComplete fail with status ",status);
            return status;
        }

        waitTime = endTime - new Date().getTime();
        status = this.waitForPendingRequests(waitTime, 3000);
        if (status != WaitStatus.StatusOK) {
            getLogger().debug("waitForPendingRequests fail with status " + status);
            return status;
        }

        getLogger().debug("WaitForPageComplete complete successfully");
        return WaitStatus.StatusOK;
    },

    getLastReqEndTimestamp: function(){
        var lastReqEndTimestamp = 0;
        for (var i = 0; i < this.browsers.length; i++) {
            if (this.browsers[i].lastReqEndTimestamp > lastReqEndTimestamp) {
                lastReqEndTimestamp = this.browsers[i].lastReqEndTimestamp;
            }
        }
        return lastReqEndTimestamp;
    },

    waitForPendingRequests: function(maxWaitTime, maxReadGap) {
        getLogger().debug("WaitForPendingRequests: " + maxWaitTime + ", maxReadGap of "+ maxReadGap);
        if (this.getPendingRequests() <= 0) {
            getLogger().debug("Without pending request, needn't wait for pending request");
            return WaitStatus.StatusOK;
        }

        var timeoutRemaining = maxWaitTime;
        var main = Cc["@mozilla.org/thread-manager;1"].getService(Ci.nsIThreadManager).mainThread;
        var startTime = new Date().getTime();
        var hadError = false;
        while ((this.getPendingRequests() > 0 || currentTime - this.getLastReqEndTimestamp() < 1000 ) &&
            timeoutRemaining > 0 && !hadError) {
            this.doBusyWait(WAIT_SLICE);

            timeoutRemaining -= WAIT_SLICE;
            hadError = this.hadInternalError();
            var currentTime = new Date().getTime();
            var readGap = currentTime - this.lastDataRead;
            var pendingTime = currentTime - startTime;
            if ((readGap > maxReadGap) && (pendingTime > maxReadGap))  {
                getLogger().debug( "Remaining " + timeoutRemaining + "ms, read wait exceeded gap = " + readGap);
                break;
            }
        }

        if (hadError){
            return WaitStatus.StatusError;
        }else{
            return (timeoutRemaining <= 0) ? WaitStatus.StatusTimeout: WaitStatus.StatusOK;
        }
    },

    waitAfterPageComplete: function(waitTime) {
        if (WAIT_SLICE < waitTime) {
            getLogger().info("Wait after Page Complete for " + WAIT_SLICE);
            var status = this.waitForTime(WAIT_SLICE);
            return status ? WaitStatus.StatusOK : WaitStatus.StatusError;
        } else {
            getLogger().warn("Delay of " + WAIT_SLICE + " exceeds max waitTime, returning timeout.");
            return WaitStatus.StatusTimeout;
        }
    },

	getPendingRequests: function() {
        var pendingRequests = 0;
        for (var i=0; i < this.browsers.length; i++) {
            pendingRequests += this.browsers[i].pendingRequests.size();
        }
        return pendingRequests;
    }
}

function BrowserMonitor(browser, recorder) {
    this.timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
    this.browser = browser;
    this.timeoutExpired = false;
    this.pageCompleted = false;
    this.pendingDocNum = 0;
    this.waitingForPageComplete = false;
    this.pendingRequests = new RequestSet();
    this.waitingForNetComplete = false;
    this.lastActiveTimestamp = 0;
    this.recorder = recorder;
    this.instrument();
}

BrowserMonitor.prototype = {
    instrument: function() {
        this.browser.addProgressListener(this, Ci.nsIWebProgress.NOTIFY_ALL);
        activityDistributor.addObserver(this);
    },

    uninstrument: function() {
        this.browser.removeProgressListener(this);
        activityDistributor.removeObserver(this);
    },

    doBusyWait: function(waitTime){
        if (!waitTime) return;
        this.busyWait = true;
        this.timer.init(this, waitTime, Ci.nsITimer.TYPE_ONE_SHOT);
        var thread = Cc["@mozilla.org/thread-manager;1"].getService(Ci.nsIThreadManager).currentThread;
        while (this.busyWait){
            thread.processNextEvent(true);
        }
    },

    waitForPageComplete: function(maxWaitTime) {
        if (this.waitingForPageComplete) {
            throw "We are already waiting for PAGE on this browser!";
        }

        if (maxWaitTime == undefined) {
            throw "Unspecified maxWaitTime!";
        }

        if (this.pageCompleted) {
            getLogger().debug("waitForPageComplete: already complete");
            return WaitStatus.StatusOK;
        }

        getLogger().debug("start waiting for PAGE for ", maxWaitTime, "ms");

        this.waitingForPageComplete = true;
        this.timeoutExpired = false;
        // start the timer if necessary
        if (maxWaitTime > 0) {
            this.timer.init(this, maxWaitTime, Ci.nsITimer.TYPE_ONE_SHOT);
        } else if (maxWaitTime <= 0) {
            this.timeoutExpired = true;
        }

        // wait for page complete until timeout expired
        var main = Cc["@mozilla.org/thread-manager;1"].getService(Ci.nsIThreadManager).mainThread;
        var hadError = false;
        while (!this.timeoutExpired && !this.pageCompleted && !hadError){
            main.processNextEvent(true);
            if(this.recorder){
                hadError = this.recorder.hadInternalError();
            }
        }

        this.timer.cancel();
        this.pageCompleted = true;
        this.waitingForPageComplete = false;

        getLogger().debug("wait PAGE completed because " + (this.timeoutExpired ? "timeout expired" : "page completed") + (hadError ? ", hadError!" : ""));

        if (hadError){
            return WaitStatus.StatusError;
        }else{
            return (this.timeoutExpired) ? WaitStatus.StatusTimeout: WaitStatus.StatusOK;
        }
    },

    waitForNetRequestsComplete: function(maxWaitTime, maxIdleTime) {
        if (this.waitingForNetComplete) {
            throw "We are already waiting for NET on this browser!";
        }

        if (maxWaitTime == undefined) {
            throw "Unspecified maxWaitTime!";
        }

        getLogger().debug("start waiting for NET for", maxWaitTime, "ms");

        this.lastActiveTimestamp = new Date().getTime();
        this.waitingForNetComplete = true;
        var netCompleted = false;
        var hadError = false;
        while (maxWaitTime > 0 && !netCompleted && !hadError) {
            this.doBusyWait(WAIT_SLICE);
            var idleTime = new Date().getTime() - this.lastActiveTimestamp;
            netCompleted = this.pendingRequests.isEmpty() && (idleTime >= maxIdleTime);

            if (this.recorder) {
                hadError = this.recorder.hadInternalError();
            }

            maxWaitTime = maxWaitTime - WAIT_SLICE;
        }

        this.waitingForNetComplete = false;

        if (hadError) {
            return WaitStatus.StatusError;
        } else {
            return (maxWaitTime <= 0) ? WaitStatus.StatusTimeout: WaitStatus.StatusOK;
        }
    },

    safeGetName: function(request) {
        return request ? request.name : "request was null";
    },

    stop: function() {
        if (this.browser.stop) {
            this.browser.stop(Ci.nsIWebNavigation.STOP_ALL);
        }
    },

    // nsISupports
    QueryInterface: function(iid) {
        if (iid.equals(Ci.nsIWebProgressListener) ||
            iid.equals(Ci.nsISupportsWeakReference) ||
            iid.equals(Ci.nsIObserver) ||
            iid.equals(Ci.nsISupports))
                return this;

        throw Cr.NS_ERROR_NO_INTERFACE;
    },

    // nsIObserver
    observe: function(subject, topic, data) {
        try{
            if (topic == "timer-callback") {
                this.timeoutExpired = true;
                this.busyWait = false;
                this.timer.cancel();
            }
        }catch(ex) {
            logException(ex,"observe");
        }
    },

    /* nsIActivityObserver */
    observeActivity: function(httpChannel, activityType, activitySubtype, timestamp, extraSizeData, extraStringData) {
        try {
            if (httpChannel instanceof Ci.nsIHttpChannel) {
                if (activityType == nsIHttpActivityObserver.ACTIVITY_TYPE_HTTP_TRANSACTION) {
                    if (activitySubtype == nsIHttpActivityObserver.ACTIVITY_SUBTYPE_REQUEST_HEADER) {
                        this.lastActiveTimestamp = new Date().getTime();
                        var request = httpChannel.QueryInterface(Ci.nsIRequest);
                        this.pendingRequests.add(request);
                        var newListener = new TracingListener(this.recorder);
                        httpChannel.QueryInterface(Ci.nsITraceableChannel);
                        newListener.originalListener = httpChannel.setNewListener(newListener);
                    } else if (activitySubtype == nsIHttpActivityObserver.ACTIVITY_SUBTYPE_TRANSACTION_CLOSE) {
                        this.lastReqEndTimestamp = this.lastActiveTimestamp = new Date().getTime();
                        var request = httpChannel.QueryInterface(Ci.nsIRequest);
                        this.pendingRequests.remove(request);
                    } else if (activitySubtype == nsIHttpActivityObserver.ACTIVITY_SUBTYPE_RESPONSE_COMPLETE) {
                        this.lastActiveTimestamp = new Date().getTime();
                    }
                }
            }
        } catch (ex) {
            logException(ex,"observeActivity");
        }
    },

    onStateChange: function(webProgress, request, stateFlags, status) {
        var requestName = this.safeGetName(request);
        if(/((chrome|resource|wyciwyg):\/\/)|about:/.test(requestName)) {
            return;
        }

        const wp = Ci.nsIWebProgressListener;
        var isRequest = ((stateFlags & wp.STATE_IS_REQUEST) != 0);
        var isDocument = ((stateFlags & wp.STATE_IS_DOCUMENT) != 0);
        var isNetwork = ((stateFlags & wp.STATE_IS_NETWORK) != 0);
        var isWindow = ((stateFlags & wp.STATE_IS_WINDOW) != 0);
        var isStart = ((stateFlags & wp.STATE_START) != 0);
        var isRedirecting = ((stateFlags & wp.STATE_REDIRECTING) != 0);
        var isTransferring = ((stateFlags & wp.STATE_TRANSFERRING) != 0);
        var isNegotiating = ((stateFlags & wp.STATE_NEGOTIATING) != 0);
        var isStop = ((stateFlags & wp.STATE_STOP) != 0);

        var state = "WP " + requestName;
        if(isRequest) state += " request";
        if(isDocument) state += " document";
        if(isNetwork) state += " network";
        if(isWindow) state += " window";
        if(isStart) state += " start";
        if(isRedirecting) state += " redirecting";
        if(isTransferring) state += " transferring";
        if(isStop) state += " stop";
        if(isStart) state = ">>> STARTED " + state;
        if(isStop) state = "<<< STOPPED " + state;

        if(isDocument && (webProgress.DOMWindow == webProgress.DOMWindow.top)) {
            if (isStart) {
                this.pendingDocNum++;
                this.pageCompleted = false;
                //getLogger().debug("Top start , page start  " + requestName + ", Pding document increased, count: " + this.pendingDocNum);
            } else if (isStop) {
                this.pendingDocNum--;
                //getLogger().debug("Top stop, page stop " + requestName + " Pending document decreased, count: "+ this.pendingDocNum);
                if (this.pendingDocNum <= 0) {
                    this.pageCompleted = true;
                }
            }
        }
    }
}

function RequestSet() {
    this.array = [];
}

RequestSet.prototype = {
    add: function(req) {
        var index = this.array.indexOf(req);
        if (index == -1) {
            this.array.push(req);
        }
    },

    remove: function(req) {
        var index = this.array.indexOf(req);
        if (index != -1) {
            this.array.splice(index, 1);
        }
    },

    clear: function() {
        this.array.length = 0;
    },

    size: function() {
        return this.array.length;
    },

    isEmpty: function() {
        if (this.array.length == 0) return true;

        var len = this.array.length;

        var vrl = 0;
        for (var i = 0; i < len; i++) {
            var request = this.array[i];
            var status = request.status;
            if (0 != status) {
                vrl++;
            }
        }
        return vrl == len;
    },

    toString: function() {
        var len = this.array.length;
        var msg = (len > 0) ? "pendingRequest length is " + len + ",they are " : "no pendingRequest";
        for (var i = 0; i < len; i++) {
            var request = this.array[i];
            var url = request ? request.name : "request was null";
            var status = request.status;
            msg += "\r\n url: " + url + ", status: " + status;
        }
        return msg;
    } , 

    visitRequest : function( callbackFunc ) {
        for (var i = 0 ; i < this.array.length ; i++) {
            callbackFunc(this.array[i]);
        }
    }
}

function TracingListener(recorder) {
    this.recorder = recorder;
}
TracingListener.prototype = {
    originalListener: null,

    onDataAvailable: function(request, context, inputStream, offset, count) {
        this.recorder.lastDataRead = new Date().getTime();
        try
        {
            if (this.originalListener) {
                this.originalListener.onDataAvailable(request, context, inputStream, offset, count);
            }
        } catch(ex) {
            logException(ex,"onDataAvailable");
        }
	},

    onStartRequest: function(request, context) {
        try {
            if (this.originalListener) {
                this.originalListener.onStartRequest(request, context);
            }
        } catch(ex) {
            logException(ex,"onStartRequest");
        }
	},

    onStopRequest: function(request, context, statusCode ) {
        try {
            if (this.originalListener) {
                this.originalListener.onStopRequest(request, context, statusCode);
            }
        } catch(ex) {
            logException(ex,"onStopRequest");
        }
	},

	QueryInterface: function(aIID) {
		if (aIID.equals(Ci.nsIStreamListener) ||
            aIID.equals(Ci.nsISupports)) {
			return this;
		}
		throw Cr.NS_NOINTERFACE;
	}
}